/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.controlador;

import com.sun.istack.logging.Logger;

import edu.uoc.tcfj2ee_gatd.model.comu.Util;

/**
 * Generalització d'un filtre que restringeix el llistat d'un tipus
 * d'entitats.
 * <p>
 * Ofereix algunes facilitats per a muntar frases JPQL a les seves
 * descendents.
 * </p>
 * @author Jordi Tejero
 *
 */
public abstract class FiltreAbs {

	/**
	 * El logger
	 */
	private Logger logger = Logger.getLogger(this.getClass());
	
	/**
	 * "Builder" que fem servir per a muntar la clàusula SELECT
	 */
	private final StringBuilder select = new StringBuilder();
	
	/**
	 * "Builder" que fem servir per a muntar la clàusula WHERE
	 */
	private final StringBuilder where = new StringBuilder();

	/**
	 * Constructor. Crea l'objete i l'inicialitza
	 */
	public FiltreAbs() {
		super();
		/*this.getLogger().finer(this.getClass().getSimpleName() + 
		"#FiltreAbs()");*/
		this.ini();
		/*this.getLogger().finer(this.getClass().getSimpleName() + 
			"#FiltreAbs(). Filtre creat");*/
	}
	
	/**
	 * Neteja la sentència JPQL, eliminant les clàusules SELECT i WHERE.
	 * Els atributs que determinen les condicions romanen.
	 * @return <code>this</code>
	 */
	protected FiltreAbs netejaJPQL() {
		this.select.setLength(0);
		this.where.setLength(0);
		return this;
	}

	/**
	 * Dona la sentència JPQL muntada
	 * @return la concatenació de les clàusules SELECT i WHERE, separades
	 *  per un salt de línia
	 */
	protected String getSentencia() {
		if(this.where.length() == 0)
			return this.select.toString();
		// else
		return this.select.toString() + Util.LS + this.where;
	}
	
	

	/**
	 * Afegeix un fragment a la clàusula SELECT
	 * @param fragment el fragment a afegir
	 */
	protected FiltreAbs posaSelect(String fragment) {
		this.select.append(fragment);
		return this;	
	}
	
	/**
	 * Afegeix una restricció AND a la clàusula WHERE
	 * @param fragment fragment JPQL que dona la condició
	 * @return this 
	 */
	protected FiltreAbs posaAnd(CharSequence fragment) {
		if(this.where.length() == 0)
			this.where.append(Util.LS).append("WHERE ");
		else
			this.where.append(Util.LS).append("AND ");
		this.where.append("(").append(fragment).append(")");
		
		return this;
	}
	
	/**
	 * Afegiment condicionat d'una restricció AND a la clàusula WHERE,
	 * del tipus [atribut] = '[cadena]'. Si [cadena] es <code>null</code> o la 
	 * cadena buida, el mètode no té cap efecte.
	 * @param atribut el nom de l'atribut sobre el que s'imposa la condició
	 * @param cadena el valor que ha de prendre l'atribut. Si és 
	 * <code>null</code> o "", no s'afegeix cap restricció al filtre.
	 * @return <code>this</code>
	 */
	protected FiltreAbs posaAnd(String atribut, String cadena) {
		if(cadena != null && !cadena.isEmpty())
			this.posaAnd(atribut + " = '" + cadena + "'");
		return this;
	}

	/**
	 * Afegiment condicionat d'una restricció
	 * que demana igualtat d'un atribut amb un element E d'un tipus enumerat T.
	 * <p>
	 * En concret, afegeix a la clàusula WHERE de la sentència, un condició
	 * AND del tipus [atribut] = [nom canònic de T].[nom d'E].
	 * </p>
	 * Si E es <code>null</code> el mètode no té cap efecte.
	 * @param atribut el nom de l'atribut sobre el que s'imposa la condició
	 * @param element l'element de tipus enumerat, E. Si és 
	 * <code>null</code>, no s'afegeix cap restricció al filtre.
	 * @return <code>this</code>
	 */
	protected FiltreAbs posaAnd(String atribut, Enum<?> element) {
		if(element != null)
			this.posaAnd(atribut + " = " + this.getEnumStr(element));
		return this;
	}	
	
	/**
	 * Afegiment condicionat d'una restricció que demana desigualtat d'un 
	 * atribut amb un element E d'un tipus enumerat T.
	 * <p>
	 * En concret, afegeix a la clàusula WHERE de la sentència, un condició
	 * AND del tipus [atribut] <> [nom canònic de T].[nom d'E].
	 * </p>
	 * Si E es <code>null</code> el mètode no té cap efecte.
	 * @param atribut el nom de l'atribut sobre el que s'imposa la condició
	 * @param element l'element de tipus enumerat, E. Si és 
	 * <code>null</code>, no s'afegeix cap restricció al filtre.
	 * @return <code>this</code>
	 */
	protected FiltreAbs posaAndNo(String atribut, Enum<?> element) {
		if(element != null)
			this.posaAnd(atribut + " <> " + this.getEnumStr(element));
		return this;
	}		
	
	/**
	 * Dona la cadena que representa un element de tipus enumerat E,
	 * amb el nom de la seva classe, T, i el del element.
	 * <p>
	 * Útil per a muntar condicions JPQL.
	 * </p>
	 * @param element el element de tipus enumerat E
	 * @return cadena del tipus [nom canònic de T].[nom d'E]
	 */
	protected String getEnumStr(Enum<?> element) {
		return element == null? null :
			element.getClass().getCanonicalName() + "." + element.name()
		;
	}
	
	/**
	 * Afegit condicionat d'una restricció que demana que un atribut
	 * prengui un valor entre un seguit de valors cadena.
	 * <p>
	 * Si <code>atribut == null</code> o <code>valors == null</code>, no 
	 * afegeix cap restricció al filtre.
	 * </p>
	 * @param atribut el nom de l'atribut
	 * @param valors el seguit de valors cadena
	 * @return <code>this</code>
	 */
	protected FiltreAbs posaAndIn(String atribut, Iterable<String> valors) {
		
		if(atribut == null || valors == null)
			return this;
		
		StringBuilder fragment = new StringBuilder();
				
		for(String v : valors) {
			if(fragment.length() > 0)
				fragment.append(Util.LS).append("OR ");
			fragment.append(atribut).append(" = '").append(v).append("'");
		}
		
		this.posaAnd(fragment);
		
		return this;
	}
	
	/**
	 * Afegiment condicionat d'una restricció al filtre, que demana que un
	 * atribut A de text, encaixi en un patró LIKE d'SQL, P. 
	 * <p>
	 * Si P != null, afegeix a la clàusula WHERE una condició AND del tipus
	 * [A] LIKE [P].
	 * </p>
	 * <p>Si P == null o P == "", la funció no té cap efecte.
	 * </p>
	 * @param atribut el nom, A, de l'atribut que ha d'encaixar en el patró
	 * @param patro el patró LIKE, P.
	 * @return <code>this</code>
	 */
	protected FiltreAbs posaAndLike(String atribut, String patro) {
		if(patro != null && !patro.isEmpty())
			this.posaAnd(atribut + " LIKE '" + patro + "'");
		
		return this;
	}
	
	
	/**
	 * Accessor de lectura de l'atribut logger
	 * @return el valor de l'atribut <code>logger</code>
	 */
	protected Logger getLogger() {
		return this.logger;
	}

	/**
	 * Neteja el filtre, cridant la funció d'inicialització. Els
	 * atributs de la subclasse que determinen les condicions que
	 * imposa el filtre, hauràn estat anul·lats.
	 */
	public void neteja() {
		this.ini();
	}
	
	/**
	 * A implementar per les classes descents, donarà la frase JPQL
	 * corresponent a l'aplicació del filtre al conjunt de tots els
	 * usuaris de la classe
	 * @return sentència JPQL que troba els usuaris que passen el filtre
	 */
	public abstract String getJPQL();
	
	/**
	 * Funció d'inialització, implementada per les classes descendents.
	 * <p>
	 * Invocada pel constructor i per la funció de neteja del filtre
	 * </p>
	 */
	protected abstract void ini();
	
}