/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.controlador.peticions;

import java.util.List;
import java.util.logging.Level;

import javax.annotation.PostConstruct;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.ManagedProperty;
import javax.faces.bean.RequestScoped;

import edu.uoc.tcfj2ee_gatd.controlador.EJBControlador;
import edu.uoc.tcfj2ee_gatd.controlador.usuaris.ControladorSessio;
import edu.uoc.tcfj2ee_gatd.model.comu.Util;
import edu.uoc.tcfj2ee_gatd.model.peticions.GATDExcepcioPeticions;
import edu.uoc.tcfj2ee_gatd.model.peticions.GestorPeticions;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Client;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Peticio;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Queixa;
import edu.uoc.tcfj2ee_gatd.model.usuaris.GestorUsuaris;


/**
 * Controlador MVC, responsable de la gestió de les peticions
 * @author Jordi Tejero
 *
 */
@ManagedBean(name = "peticions")
@RolesAllowed({"Client", "Tecnic", "Controlador"})
@RequestScoped
public class ControladorPeticions 
extends EJBControlador {
	
	/**
	 * El gestor de peticions
	 */
	@EJB
	private GestorPeticions gPeticions;
	
	/**
	 * Gestor d'usuaris
	 */
	@EJB
	private GestorUsuaris gUsuaris;
	
	/**
	 * El controlador de sessió
	 */
	@ManagedProperty(value = "#{sessio}")
	private ControladorSessio sessio;

	
	/**
	 * El filtre que dona el llistat de les peticions
	 */
	private FiltrePeticions filtrePeticions;	
	
	/**
	 * La petició actual. És la que es crea o edita
	 */
	private Peticio peti;
	
	/**
	 * Possible queixa sobre la petició actual. Si el usuari ho indica,
	 * l'enllaçarem a la petició.
	 */
	private Queixa queixa;
	
	/**
	 * Llista de peticions d'un usuari que es mostra
	 */
	private List<Peticio> peticions;
	
	/**
	 * Constructor. Inicialitza els camps.
	 */
	public ControladorPeticions() {
		super();
		this.setPeti(new Peticio());
		this.queixa = new Queixa();
		this.filtrePeticions = new FiltrePeticions();
	}
	
	/**
	 * Incialització. Si la petició HTTP tenia un paràmetre amb la referència
	 * de la petició, la carreguem.
	 * <p>
	 * Matusser, però funciona
	 * </p>
	 * 
	 */
	@PostConstruct
	public void ini() {
		// Traiem la referència directament de la petició http
		String ref = this.sessio.getParamHttp(".*refPeti");
		try {
			if(ref != null)
				this.setPeti(this.gPeticions.cerca(Long.parseLong(ref)));
		}
		catch(Throwable t) {
			this.getLogger().log(Level.SEVERE, 
				"*** ERROR INESPERAT a " + this.getClass().getSimpleName() + "#ini()", 
				t);
		}
	}
	
	
	/**
	 * Permet l'injecció del controlador de sessió
	 * @param sessio el nou valor de l'atribut sessió
	 */
	public void setSessio(ControladorSessio sessio) {
	    this.sessio = sessio;
	}

	/**
	 * Accessor de lectura de l'atribut filtre
	 * @return el valor de l'atribut <code>filtre</code>
	 */
	public FiltrePeticions getFiltre() {
		return this.filtrePeticions;
	}



	/**
	 * Establiment de la petició actual a partir de la seva referència.
	 * <p>
	 * Cerca la petició al gestor de persistència i actualitza en
	 * conseqüència <code>this.peti</code> i <code>this.queixa</code>.
	 * </p>
	 * @param referencia la referència de la petició
	 */
	public void setRefPeti(Long referencia) {
		if(referencia != null) try {
			Peticio p = this.gPeticions.cerca(referencia);
			assert p != null;
			this.setPeti(p);
			
			this.getLogger().finer(this.getClass().getSimpleName() + 
			"#setRefPeti(" + referencia + "). " +
			"Estat " + this.peti.getEstat().name() + ". " +
			"Versió " + this.peti.getVersio());
						
		}	
		catch (Exception e) {
			this.getLogger().log(Level.FINE, 
				"ControladorPeticions#setRefPeti(" + referencia + ")", e.toString());
				this.pintaError(this.getMsg("err_excp_Inesperat", e.toString()));
		}
		
	}	
	
	/**
	 * Publica la referència de la petició actual
	 * @return la referència de la petició actual {@link #peti}, o
	 *  <code>null</code> si no és establerta
	 */
	public Long getRefPeti() {
		return this.peti == null? null : this.peti.getReferencia();
	}
		
	/**
	 * Accessor de lectura de la petició que s'edita
	 * @return el valor de l'atribut <code>peti</code>
	 */
	public Peticio getPeti() {
		if(this.peti == null) {
			this.setPeti(new Peticio());
		}
		return this.peti;
	}

	/**
	 * Estableix la petició actual i la queixa que pugui tenir.
	 * <p>
	 * <code>this</code> es registra com observador de l'entitat 
	 * <code>Peticio</code>.
	 * </p>
	 * @param peti la nova petició actual
	 */
	private void setPeti(Peticio peti) {
		this.peti = peti;		
		this.queixa = 
			this.peti.getQueixa() == null? new Queixa() : peti.getQueixa();
	}
	

	/**
	 * Publica la queixa sobre la petició actual
	 * @return el valor de l'atribut <code>queixa</code>
	 */
	public Queixa getQueixa() {
		return this.queixa;
	}

	/**
	 * Indica si hi ha una queixa sobre la petició, no anul·lada
	 * @return <code>true</code> si hi ha una
	 */
	public boolean getQueixada() {
		return this.peti.getQueixa() != null && !this.peti.getQueixa().isBaixa();
	}
	
	/**
	 * Assigna la queixa a la petició, o anul·la la que tenia.
	 * @param queixada <code>true</code> per a indicar que la queixa
	 *  ha de ser assignada a la petició actual.
	 */
	public void setQueixada(boolean queixada) {
		this.getLogger().finer(
			"ControladorPeticions#setQueixada(" + queixada + ")");
		if(queixada) {
			this.queixa.alta();
			this.queixa.setAutor(this.sessio.getUsuari());
			this.peti.setQueixa(this.queixa);
		}
		else 
			this.queixa.baixa();			
	}		
	

	/**
	 * Navegació web. Aplica el filtre al llistat de peticions i recarrega
	 * la pàgina actual.
	 * @return <code>null</code>
	 */
	public String filtra() {
		this.peticions = this.gPeticions.llista(this.filtrePeticions.getJPQL());
		return null;
	}
	
	/**
	 * Navegació web. Elimina totes les restriccions del filtre i 
	 * carrega la pàgina actual amb el llistat de totes les peticions.
	 * @return <code>null</code>
	 */
	public String netejaFiltre() {
		this.getFiltre().neteja();
		return this.filtra();
	}
	
	/**
	 * Dona la llista de totes les peticions pendents d'un client. Si no
	 * existeix, la crea.
	 * @param nomClient el nom d'usuari del client
	 * @return la llista de peticions
	 * @throws ClassCastException si <code>cl</code> no és un client
	 */
	public List<Peticio> getPeticions(String nomClient) {
		assert nomClient != null;		
		assert this.filtrePeticions != null;
		
		if(this.peticions == null 
			|| !nomClient.equals(this.filtrePeticions.getNomClient())) {
			this.filtrePeticions.setNomClient(nomClient);
			this.peticions = this.gPeticions.llista(this.filtrePeticions.getJPQL());
		}
			
	    return this.peticions;
	}
	
	/**
	 * Dona la llista de les peticions que passen el filtre actual. Si no
	 * existeix, la crea.
	 * @return la llista de les peticions que passen el filtre
	 */
	public List<Peticio> getPeticions() {
		assert this.filtrePeticions != null;
		if(this.peticions == null)  try {			
			this.peticions = this.gPeticions.llista(this.filtrePeticions.getJPQL());
		}
	    catch (Exception e) {
			this.getLogger().log(Level.FINE, 
					"ControladorPeticions#getPeticions()", e.toString());
				this.pintaError(this.getMsg("err_excp_Inesperat", e.toString()));
		}
			
	    return this.peticions;		
	}

	
	/**
	 * Crea una nova petició del client que s'indiqui
	 * @param nomClient nom d'usuari del client que fa la petició
	 */
	public void nova(String nomClient) {
		Client cl = this.gUsuaris.cerca(Client.class, nomClient);
		assert cl != null;
		this.setPeti(new Peticio(cl));
	}
	
	
	/**
	 * Alta de la nova petició, associant-la al client que s'indica.
	 * Se'n verifica l'adreça i s'hi assigna a un tècnic.
	 * <p>
	 * Cas d'error, escriu un missatge
	 * </p>
	 * 
	 * @param nomClient el nom del client de la petició
	 * 
	 * @return La URL del llistat de peticions si ha pogut donar d'alta la
	 *  petició, <code>null</code> si no es així
	 */
	public String alta(String nomClient) {
		boolean exit = false;
		this.getLogger().fine("ControladorPeticions#alta()");
		Client client = this.gUsuaris.cerca(Client.class, nomClient);
		assert client != null;
		
		this.peti.setClient(client);
		try {
			this.gPeticions.novaPeticio(this.peti);
			exit = true;
			this.pintaInfo("fp_InfoCreada", "misg_Exit", this.peti.getReferencia());
		} 
		catch (GATDExcepcioPeticions e) {
			this.getLogger().log(Level.FINE, 
				"ControladorPeticions#alta(), alta de: " 
				+ Util.LS + this.peti, e);
			this.pintaError(e);
		}
		catch (Exception e) {
			this.getLogger().log(Level.FINE, 
					"ControladorPeticions#alta(), alta de: " 
					+ Util.LS + this.peti, e);
				this.pintaError(this.getMsg("err_excp_Inesperat", e.toString()));
		}
		return exit? "administraPeticions" : null;
	}
	
	/**
	 * Acceptar canvis a la petició actual
	 * @return <code>null</code>
	 */
	public String accepta() {
	    this.getLogger().finer(
	           "ControladorPeticions#accepta(). " +
	           "peti.referencia = " + this.peti.getReferencia()); 	    	   
	    try {
		    this.setPeti(this.gPeticions.guarda(this.peti));
		    this.pintaInfo("fp_InfoModificada", "misg_Exit", 
		            this.peti.getReferencia());
	    } catch (GATDExcepcioPeticions e) {
			this.pintaError(e);
		}
	    catch (Exception e) {
			this.getLogger().log(Level.FINE, 
					"ControladorPeticions#accepta()" + this.peti, e);
				this.pintaError(this.getMsg("err_excp_Inesperat", e.toString()));
		}
	        
	    this.getLogger().finer(
	    	"ControladorPeticions#accepta(). Acceptat canvis " + this.peti);
	    
		//return "administraPeticions";
		return null;
	}
	
	/**
	 * Anul·la la petició actual
	 * @return <code>null</code>
	 */
	public String anulla() {
	    this.getLogger().finer(
           "ControladorPeticions#anulla(). " +
           "peti.referencia = " + this.peti.getReferencia()); 		
	    
	    try {
			this.gPeticions.anulla(this.peti);
		} catch (GATDExcepcioPeticions e) {
			this.pintaError(e);
		}
	    catch (Exception e) {
			this.getLogger().log(Level.FINE, 
					"ControladorPeticions#accepta()" + this.peti, e);
				this.pintaError(this.getMsg("err_excp_Inesperat", e.toString()));
		}
	    	    
		return null;
	}
	

}
