/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.controlador.usuaris;

import java.util.LinkedList;
import java.util.List;

import javax.annotation.security.RolesAllowed;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.RequestScoped;
import javax.faces.model.SelectItem;

import edu.uoc.tcfj2ee_gatd.model.atencio.entitats.Tecnic;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Seu;
import edu.uoc.tcfj2ee_gatd.model.usuaris.GATDExcepcioUsuaris;

/**
 * Gestió dels tècnics per part dels controladors. Navega entre la pàgina
 * del llistat i la fitxa d'un tècnic.
 * <p>
 * Afegeix al comportament de la classe base, la navegació entre pàgines JSF i
 * la gestió de les marques. 
 * </p>
 * @author Jordi Tejero
 *
 */
@ManagedBean(name = "tecnics")
@RolesAllowed({"Controlador"})
@RequestScoped
public class ControladorTecnics extends ControladorUsuaris<Tecnic> {

	/**
	 * Nom de l'atribut HTML de la llista de tècnics marcats
	 */
	private static final String MARCATS = "marcats";
	
	/**
	 * El filtre que dona el llistat dels tècnics.
	 */
	private FiltreTecnics filtreTecnics;
	
	/**
	 * Codi de la nova seu del tècnic, quan es canvia
	 */
	private String codiNovaSeu;
	
	
	/**
	 * Constructor
	 *
	 * @throws GATDExcepcioUsuaris
	 */
	public ControladorTecnics() throws GATDExcepcioUsuaris {
		super(Tecnic.class);
		this.filtreTecnics = new FiltreTecnics();
		/*this.getLogger().finer(this.getClass().getSimpleName() + 
				"#ControladorTecnics()");*/
	}


	/**
	 * Publica el filtres del llistat de tècnics
	 * @return el valor de l'atribut <code>filtreTecnics</code>
	 */
	@Override
	public FiltreTecnics getFiltre() {
		return this.filtreTecnics;
	}
	


	/**
	 * Publica el nom de l'atribut HTML que ha de donar els noms
	 * dels tècnics marcats.
	 * <p>
	 * Per a us de les pàgines JSF.
	 * </p>
	 * @return el valor de l'atribut <code>MARCATS</code>
	 */
	public String getMARCATS() {
		return MARCATS;
	}

	/**
	 * Publica el codi de la seu del tècnic
	 * @return el valor de l'atribut codiSeu
	 */
	public String getCodiSeu() {
		return this.getActual() == null? null :
			this.getActual().getSeu().getCodi();
	}
	

	/**
	 * Modifica el codi de la nova seu del tècnic, evitant de posar-hi
	 * un valor <code>null</code>
	 * @param codiSeu el nou valor de l'atribut codiSeu
	 */
	public void setCodiSeu(String codiSeu) {
		this.codiNovaSeu = codiSeu;
	}


	/* Especialització de la validació d'usuari per a Tecnic.
	 * Demanem la validesa com a usuari i, a més a més, te tingui e-mail i
	 * seu. 
	 * Canvia si cal la seu del tècnic, controlant també que no es canviï de seu 
	 * un que estigui actiu.
	 * @see edu.uoc.tcfj2ee_gatd.controlador.usuaris.ControladorUsuaris#validaActual()
	 */
	@Override
	protected boolean validaActual() {
		
		boolean valid = super.validaActual();		
		
		if(this.getActual().getContacte().geteMail() == null 
		|| this.getActual().getContacte().geteMail().trim().isEmpty()) {
			this.pintaError(super.getMsg("de_CalEMail"));
			valid = false;
		}
		
		if(this.codiNovaSeu != null 
		&& !this.codiNovaSeu.equals(this.getActual().getSeu().getCodi())) {
			if(this.getActual().isActiu())
				this.pintaError(super.getMsg("usua_err_SeuTecnicActiu", 
					this.getActual().getNomUsuari()));
			else {
				Seu seu = this.gUsuaris.cercaSeu(this.codiNovaSeu);
				if(seu == null) 
					this.pintaError(super.getMsg("usua_err_SeuNoExisteix", 
						this.codiNovaSeu));
				else
					this.getActual().setSeu(seu);
			}
				
		}
			
			
		
		if(this.getActual().getSeu() == null
		|| this.getActual().getSeu().getCodi() == null) {
			this.pintaError(super.getMsg("de_CalSeu"));
			valid = false;
		}
		
		this.getLogger().finer(this.getClass().getSimpleName() + 
		"#validaActual() -> " + valid);
		
		return valid;
	}

	/**
	 * Navegació web segons els cas d'ús "Modifica Situació Tècnics". 
	 * <p>Commuta la situació d'activitat de tots els
	 * tècnics marcats a la petició HTTP, i es queda a la mateixa pàgina.
	 * </p>
	 * <p>
	 * Per a evitar actuar sobre tècnics que puguin haver estat modificats
	 * (per exemple, donats de baixa o activats/desactivats) després d'haver-se
	 * carregat el llistat, els paràmetres de la petició HTTP contenen,
	 * a més dels noms d'usuari dels tècnics, també els números de versió
	 * d'entitat de cada un.
	 * </p>
	 * <p>
	 * <b>pre</b> S'ha fet una petició HTTP amb paràmtres de nom {@link #MARCATS}
	 * i valors de la forma [nom tècnic]:[número versió].
	 *  
	 * </p>
	 * <p>
	 * <b>post</b> Una de dos:
	 * <ol>
	 * 	<li>Els tècnics actius d'entre els marcats han passat a inactius i
	 *  els inactius han passat a actius i les peticions pendents dels primers
	 *  han estat reassignades als segons.
	 *  </li>
	 *  <li>S'ha produït una excepció i l'estat del sistema no ha variat
	 *  </li> 
	 * </ol>
	 * </p>
	 * @return <code>null</code>
	 */
	public String modificaSituacioTecnics() {
		
		String noms[] = null;
		long versions[] = null;
		
		/*
		 * Obtenim els noms dels tècnics que volem commutar l'activitat i
		 * els seus números de versió d'entitat
		 */
		String pars[] = this.getSessio().getParamsHttp(MARCATS);
		if(pars != null) {
			noms = new String[pars.length];
			versions = new long[pars.length];
			for(int i = 0; i < pars.length; ++i) {
				String[] nv = pars[i].split(":");
				noms[i] = nv[0];
				versions[i] = Long.parseLong(nv[1]);
			}
		}
			
		
		try {
			this.gUsuaris.modificaSituacioTecnics(
				noms, versions);
		} catch (GATDExcepcioUsuaris e) {
			this.pintaError(e);
		}

		// Mostrarem sols els tècnics que permeti el filtre
		this.filtra();
		
		return null;
	}

	/**
	 * Crea una llista de <code>SelectItem</code> amb totes les seus d'alta,
	 * @return la llista creada
	 */
	public List<SelectItem> getItemsSeus() {
		
		List<SelectItem> items = new LinkedList<SelectItem>();			
		
		for(Seu s : this.gUsuaris.getSeus())  
			items.add(new SelectItem(s.getCodi(), s.getNom()));
		
		/*
		this.getLogger().finer(
			"ControladorTecnics#getItemsSeus() -> " + items.size() + " items");
		*/
		return items;			
	}
	

	
	/**
	 * Crea una llista de <code>SelectItem</code> amb totes les seus d'alta,
	 * mes un, el primer, per a indicar "totes les seus".
	 * @return la llista creada
	 */
	public List<SelectItem> getItemsSeusMesTotes() {
		List<SelectItem> items = this.getItemsSeus();
		// Als filtres, el valor null fa que no imposin la condició de Seu
		items.add(0, new SelectItem(null, this.getMsg("totes")));
		
		return items;
	}

	/**
	 * Crea una llista de <code>SelectItem</code> amb totes les seus d'alta,
	 * mes un, el primer, per a indicar que no s'ha indicat cap
	 * @return la llista creada
	 */
	public List<SelectItem> getItemsSeusMesCap() {
		List<SelectItem> items = this.getItemsSeus();
		
		items.add(0, new SelectItem(null, "------"));
		
		return items;
	}	
	
	/**
	 * Dona una cadena representant la situació d'activitat d'un tècnic,
	 * en el idioma de la sessió
	 * @param t el tècnic del que es vol representat la situació d'activitat
	 * @return cadena representant "sí" o "no" en l'idioma de la sessió, segons
	 *  si el tècnic és actiu o no.
	 */
	public String getActiuStr(Tecnic t) {
		return this.getMsg(t.isActiu()? "Si" : "No");
	}	
	
	/**
	 * Dona una cadena representant la situació d'activitat del tècnic actual,
	 * en el idioma de la sessió
	 * 
	 * @return cadena representant "sí" o "no" en l'idioma de la sessió, segons
	 *  si el tècnic actual és actiu o no.
	 */
	public String getActiuStr() {
		return this.getActiuStr(this.getActual());
	}		
	
	/**
	 * Navegació web. Dirigeix cap a la pantalla d'administració del tècnic 
	 * actual.
	 * 
	 * @return la pàgina d'administració del tècnic actual.
	 * 
	 */
	public String administra() {
		return "administraTecnic";		
	}
	
	
	/**
	 * Mira de donar d'alta el tècnic actual. Si ja hi havia un tècnic
	 * amb el mateix nom o document d'identitat, mostra un missatge d'error.
	 * <p>
	 * Navegació web: Si l'alta es pot fer, es recarrega la pàgina d'altes de 
	 * tècnics, buida. Si no, es conserven les dades entrades.
	 * </p>
	 * 
	 * @return <code>null</code>
	 */
	public String alta() {
	
		if(super.fesAlta()) 
			this.setActual(new Tecnic());
		
		return null;
		
	}

	/**
	 * Baixa del tècnic actual. Si és actiu, dona un missatge d'error.
	 * @return <code>null</code>
	 */
	@Override
	public String baixa() {
		String retorn = null;
		if(this.getActual().isActiu())
			this.pintaError(this.getMsg("usua_err_BaixaTecnicActiu", 
				this.getActual().getNomUsuari()));
		else
			retorn = super.baixa();
		
		return retorn;
	}
	
}
