/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.controlador.usuaris;

import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.regex.Pattern;

import javax.ejb.EJB;
import javax.faces.bean.ManagedProperty;
import javax.faces.model.SelectItem;

import edu.uoc.tcfj2ee_gatd.controlador.EJBControlador;
import edu.uoc.tcfj2ee_gatd.controlador.FiltreAbs;
import edu.uoc.tcfj2ee_gatd.model.comu.GATDExcepcioExecucio;
import edu.uoc.tcfj2ee_gatd.model.usuaris.GATDExcepcioUsuaris;
import edu.uoc.tcfj2ee_gatd.model.usuaris.GestorUsuaris;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.DocIdentitat;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Usuari;


/**
 * Controlador MVC genèric, que gestiona els usuaris d'una classe dada.
 * <p>
 * Conté el "bean" usuari que editen les pàgines JSF.
 * </p>
 * <p>
 * S'encarrega d'escriure els missatges a les pàgines. En particular,
 * els deguts a les excepcions.
 * </p>
 * 
 * @param <U> el tipus dels usuaris administrats
 *
 * @author Jordi Tejero
 */
public abstract class ControladorUsuaris<U extends Usuari>  
extends EJBControlador {
	
	/**
	 * El gestor d'usuaris
	 */
	@EJB
	GestorUsuaris gUsuaris;
	
	/**
	 * El controlador de la sessió actual
	 */
	@ManagedProperty(value = "#{sessio}")
	private ControladorSessio sessio;

	/**
	 * La classe dels usuaris que gestiona
	 */
	private final Class<U> classe;

	/**
	 * L'usuari actual. L'usuari que es dona d'alta o edita
	 */
	private U actual;

	/**
	 * Repetició de la clau d'accés de l'usuari actual. Per a validar-la.
	 */
	private String bisClauAcces;

	/**
	 * La llista dels usuaris que es mostra
	 */
	private List<U> llista;
	
	/**
	 * Constructor. Crea un nou usuari actual i una nova llista d'usuaris, de 
	 * la classe que s'indiqui
	 * @param classe la classe dels usuaris que gestionarà el controlador
	 * @throws GATDExcepcioUsuaris si s'ha produït un error al crear l'usuari
	 *  actual
	 */
	 ControladorUsuaris(Class<U> classe) throws GATDExcepcioUsuaris {
		this.classe = classe;
		this.setActual(this.creaUsuari());
	}
	
	/**
	 * FiltreAbs que dona la sentència SQL font de la llista d'usuaris.
	 * <p>
	 * A implementar per les classes descendents.
	 * </p>
	 * @return el filtre
	 */
	public abstract FiltreAbs getFiltre(); 
	 
	 /**
	  * Permet la injecció del controlador de sessió
	 * @param sessio el controlador de sessió
	 */
	public void setSessio(ControladorSessio sessio) {
		 this.sessio = sessio;
	}

	/**
	 * Publica el controlador de sessió, per a ús de les classes derivades 
	 * @return el valor de l'atribut <code>sessio</code>
	 */
	protected ControladorSessio getSessio() {
		return this.sessio;
	}

	/**
	 * Dona el nom d'usuari de l'actual
	 * @return el nom d'usuari de <code>this.actual</code>, o <b>null</b> si 
	 *  aquest és <code>null</code>
	 */
	public String getNomUsuari() {
		return this.actual == null? null : this.actual.getNomUsuari();
	}

	/**
	  * Estableix l'usuari actual a partir d'un nom d'usuari. Evita
	  * recarregar l'usuari actual.
	  * <p>
	  * El cerca al gestor de persistència. De trobar-lo, el referencia amb
	  * <code>this.actual</code>. De no trobar-lo, referencia un de
	  * nou.
	  * </p>
	  * <p>
	  * En qualsevol cas, estableix la copia de seva clau d'accés.
	  * </p>
	 * @param nomUsuari el nom d'usuari 
	 */
	public void setNomUsuari(String nomUsuari) {
		
		this.getLogger().finer(this.getClass().getSimpleName() + 
			"#setNomUsuari(" + nomUsuari + ")");
		// Evitem carregar dos cops el mateix usuari
		if(this.actual == null 
	    || this.actual.getNomUsuari() == null 
	    || !this.actual.getNomUsuari().equals(nomUsuari)) 
			this.carrega(nomUsuari);
	}
	
	/**
	 * Carrega un usuari pel seu nom d'idem. Si no el troba al gestor de
	 * persistència, en crea un de nou amb el nom indicat
	 * <p>
	 * En qualsevol cas, l'usuari actual ha estat modificat.
	 * </p>
	 * @param nomUsuari el nom d'usuari. Pot ser <code>null</code>, amb la
	 *  qual cosa es crearà un de nou.
	 */
	protected void carrega(String nomUsuari) {
		 this.getLogger().finer(this.getClass().getSimpleName() + 
				"#carrega(" + nomUsuari + ")");
		 U u = this.gUsuaris.cerca(this.classe, nomUsuari);
		 if(u == null) {
			 u = this.creaUsuari();
			 if(nomUsuari != null)
				 u.setNomUsuari(nomUsuari);
		 }
		 this.setActual(u);		
	}
	
	/**
	 * Recarrega l'usuari actual i resta a la mateixa pàgina
	 * @return <code>null</code>
	 */
	public String recarrega() {
		this.carrega(this.getNomUsuari());
		return null;
	}
	
	/**
	 * Crea un nou usuari de tipus <b>U</b>
	 * 
	 * @return l'usuari creat
	 * @throws GATDExcepcioExecucio si no s'ha pogut crear. Indica un error
	 *  de programació.
	 */
	private U creaUsuari() throws GATDExcepcioExecucio {
		U retorn;
		try {			
			retorn = this.classe.getConstructor().newInstance();
		} catch (Exception e) {
			this.getLogger().log(Level.SEVERE, 
				"ControladorUsuaris(" + this.classe + ")", e);
			throw new GATDExcepcioExecucio(e);
		}		
			this.getLogger().finer(
			"ControladorUsuaris#creaUsuari() -> " + retorn.getBreu());
			
		return retorn;
	}


	/**
	 * Accés a l'usuari actual.
	 * @return l'usuari actual 
	 */
	public U getActual() {
		//this.getLogger().finest("ControladorUsuaris#getActual()");
		return this.actual;
	}
	
	/**
	 * Estableix l'usuari actual i actualitza en conseqüència la
	 * repetició de la seva clau d'accés.
	 * @param usuari el nou usuari
	 */
	 void setActual(U actual) {
		assert actual != null;		 
		//this.getLogger().finer("ControladorUsuaris#setActual: " + actual);
		/*this.getLogger().finer("ControladorUsuaris#setActual: " + 
			actual.getBreu());*/
		this.actual = actual;		
		this.bisClauAcces = actual.getClauAcces();
		
	}
	 	 

	/**
	 * Publica la repetició de la clau d'accés
	 * @return el valor de l'atribut <code>bisClauAcces</code>
	 */
	public String getBisClauAcces() {
		return this.bisClauAcces;
	}

	/**
	 * Estableix la repetició de la clau d'accés
	 * @param bisClauAcces el nou valor de l'atribut bisClauAcces
	 */
	public void setBisClauAcces(String bisClauAccess) {
		this.bisClauAcces = bisClauAccess;
	}

	/**
	 * Publica la seqüència dels usuaris. Si no existeix, la crea a partir del 
	 * filtre
	 * @return <code>Iterable</code> que recorre la llista d'usuaris.
	 */
	public Iterable<U> getSequencia() {
		// this.getLogger().finer("ControladorUsuaris#getSequencia()");
		return this.getLlista();
	}
	/**
	 * Crea una llista de <code>SelectItem</code> amb tots els tipus de
	 * document d'identitad contemplats per l'aplicació.
	 * @return la llista creada
	 */
	public List<SelectItem> getItemsTipusDI() {
		
		List<SelectItem> items = new LinkedList<SelectItem>();			
		
		for(DocIdentitat.TIPUS tdi : DocIdentitat.TIPUS.values())
			items.add(new SelectItem(tdi.name(), 
				this.getMsg(tdi.name())));
		
		return items;			
	}	
	
	/**
	 * Crea una llista de <code>SelectItem</code> amb tots els tipus de
	 * document d'identitat contemplats, mes un, el primer, per a indicar 
	 * "qualsevol tipus de document d'identitat".
	 * @return la llista creada
	 */
	public List<SelectItem> getItemsTipusDIMesTots() {
		List<SelectItem> items = this.getItemsTipusDI();
		/* Als filtres, el valor null fa que no imposin la condició de
		   tipus de document d'identitat */
		items.add(0, new SelectItem(null, this.getMsg("tots")));
		
		return items;
	}	
	/**
	 * Publica la llista d'usuaris, per a ús de les classes derivades. 
	 * Si no existeix, la crea a partir del filtre
	 * @return La llista d'usuaris.
	 */
	protected List<U> getLlista() {
	    if(this.llista == null)
	        this.llista = this.gUsuaris.llista(this.getFiltre().getJPQL());
		return this.llista;
	}
	
	/**
	 * Navegació web. Aplicació de filtre i recarregar la pàgina actual.
	 * @return <code>null</code>
	 */
	public String filtra() {
		this.llista = null;
		return null;
	}
	
	/**
	 * Navegació web. Neteja el filtre i recarrega la pàgina actual amb el
	 * llistat sense filtrar. 
	 * @return <code>null</code>
	 */
	public String netejaFiltre() {
		this.getFiltre().neteja();
		return this.filtra();
	}
	

	
	/**
	 * Accepta els possibles canvis fets a l'usuari actual, actualitzant les 
	 * seves dades al gestor de persistència.
	 * <p>
	 * Si s'ha produït un conflicte d'edicions, escriu un missatge d'error.
	 * </p>
	 * <p>
	 * Si té èxit, escriu un missatge informatiu.
	 * </p>
	 * @return <code>null</code> per ha indicar que es torni ha carregar
	 *  la fitxa amb les dades actualitzades, llevat si s'ha canviat l'idioma,
	 *  en que es torna la URI de la pàgina actual, per a fer que es
	 *  regeneri.
	 */
	public String accepta() {
		
		String anar = null;
		
		this.getLogger().finer(
			"ControladorUsuaris#accepta() " + this.actual.getNomUsuari());
		
		if(this.validaActual()) try {
			this.gUsuaris.actualitza(this.actual);
			// potser hem modificar l'idioma de l'usuari de la sessió
			if(this.sessio.canviaIdiomaUsuari(this.actual))
				anar = this.sessio.pagDarrera();
			this.pintaInfo("fu_InfoModificat", "misg_Exit", 
	            this.actual.getNomUsuari());
		}
		catch(GATDExcepcioUsuaris e) {
			this.pintaError(e);
		}
		
		
		//return null;
		
		// Forcem la recàrrega de l'usuari
		this.recarrega();		
		return anar;
	}
	
	/**
	 * Valida les dades de l'usuari actual. Si no son vàlides, afegeix
	 * missatges d'error als components JSF corresponents.
	 * <p>
	 * Verifica que:
	 * <ol>
	 * <li>
	 *  El nom d'usuari i la clau fan servir sols caràcters del conjunt
	 *  [a-zA-Z_0-9#]. Altres donen problemes per a fer "login".
	 *  </li>
	 * <li>
	 * La clau d'accés coincideix amb la seva repetició.
	 * </li>
	 * </ol>
	 *
	 * @return <code>true</code> si i sols les dades entrades passen les
	 * verficacions anteriors.
	 */
	protected boolean validaActual() {
		assert this.actual != null;
		assert this.getNomUsuari() != null;
		assert this.getBisClauAcces() != null;
		
		this.getLogger().finer(this.getClass().getSimpleName() + 
			"#validaActual(). Usuari " + this.actual.getBreu());
		
		boolean valid = true; 
			
		Pattern ptrValid = Pattern.compile("[a-zA-Z_0-9# ]+");
		
		if(!ptrValid.matcher(this.getActual().getNomUsuari()).matches()
		|| !ptrValid.matcher(this.getActual().getClauAcces()).matches()) {
			valid = false;
			this.pintaError(super.getMsg("fu_err_Caracters"));
		}
			
		
		if(!this.actual.getClauAcces().equals(this.getBisClauAcces())) {
			valid = false;
			this.pintaError(super.getMsg("fu_err_BisClau"));
		}
		
		return valid;
	}

	/**
	 * Mira de donar d'alta l'usuari actual. Si ja hi havia un usuar
	 * amb el mateix nom o document d'identitat, mostra un missatge d'error 
	 * i retorna <code>null</code>.
	 * 
	 * @return <code>true</code> si i sols si l'usuari s'ha donat d'alta
	 *  normalment
	 */
	protected boolean fesAlta() {
		boolean exit = this.validaActual();
		if(exit) try {
			this.gUsuaris.alta(this.actual);
			this.pintaInfo("usua_info_UsuariCreat", "misg_Exit", 
					this.actual.getNomUsuari());
		} catch (GATDExcepcioUsuaris e) {
			exit = false;
			this.pintaError(e);
		}
		this.getLogger().finer(
			"ControladorUsuaris#alta(" + this.actual.getNomUsuari() + ") -> " +
			exit);
		/*
		if(exit)
			this.pintaInfo("usua_info_UsuariCreat", "misg_Exit");	
		*/
		return exit;
	}
	
	/**
	 * Dona de baixa l'usuari actual com a donat de baixa i escriu un missatge 
	 * amb el resultat de l'operació
	 * @return <code>null</code>, per a la romandre a la mateixa pàgina, tret si
	 *  el usuari donat de baixa és el de la sessió, que va a la
	 *  pàgina de comiat.
	 */
	public String baixa() {
		assert this.actual != null;
		
		this.getLogger().finest(
			"ControladorUsuaris#baixa(" + this.actual.getNomUsuari() + ")");		
		try {
			this.gUsuaris.baixa(this.actual);
			this.pintaInfo("fu_InfoBaixa", "misg_Exit", 
		            this.actual.getNomUsuari());			
		} catch (GATDExcepcioUsuaris e) {
			this.pintaError(e);
		}
		
		this.recarrega();
		
		// Suïcidi de l'usuari?
		if(this.actual != null && this.actual.equals(this.sessio.getUsuari()))
			return this.sessio.logOut();
		return null;
	}

	
}
