/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.controlador.usuaris;

import java.io.Serializable;

import edu.uoc.tcfj2ee_gatd.controlador.FiltreAbs;
import edu.uoc.tcfj2ee_gatd.model.comu.Util;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.DocIdentitat;

/**
 * Filtre que dona els llistats de clients
 * @author Jordi Tejero
 *
 */
public class FiltreClients extends FiltreAbs implements Serializable {

	/**
	 * Versió de serialització
	 */
	private static final long serialVersionUID = 1L;
	
	/**
	 * Tipus del document d'identitat que es cerca. Pot ser <code>null</code>,
	 * amb la qual cosa el filtre no aplica cap resticció per aquest motiu
	 */
	private DocIdentitat.TIPUS tipusDI;
	
	/**
	 * Número del document d'identitat a cercar. Si és <code>null</code>,
	 * el filtre deixa passar tots els números de document.
	 */
	private String numDI;
	
	/**
	 * Nom d'usuari del client a cercar. Si és <code>null</code>,
	 * el filtre deixa passar tots els noms d'usuari.
	 */
	private String nomUsuari;
	
	/**
	 * Patró LIKE dels noms personals dels clients a cercar. Si és 
	 * <code>null</code>, el filtre deixa passar tots els noms personals.
	 */
	private String patroNomComplet;
	
	/**
	 * Constructor. El filtre es netejat amb <code>ini()</code> i no
	 * imposa cap restricció.
	 */
	public FiltreClients() {
		super();
	}

	/* Neteja el filtre, anul·lant tots els atributs que imposen les
	 * condicions.
	 *  
	 * @see edu.uoc.tcfj2ee_gatd.controlador.FiltreAbs#ini()
	 */
	@Override
	protected void ini() {
		this.tipusDI = null;
		this.numDI = null;
		this.nomUsuari = null;
		this.patroNomComplet = null;
	}

	/**
	 * Publica el nom del tipus de document d'identitat que cerca el
	 * filtre
	 * @return el nom de l'atribut tipusDI, o <code>null</code> si el
	 *  tipus és <code>null</code>
	 */
	public String getTipusDI() {
		return this.tipusDI == null? null : this.tipusDI.name();
	}

	/**
	 * Estableix el tipus de document d'identitat
	 * @param nomTipus el nom del tipus de document. Pot ser <code>null</code>
	 * o la cadena buida "", amb la qual cosa s'anul·la el tipus de document a 
	 * cercar.
	 * 
	 * @throws IllegalArgumentException si [nomTipus] no él el nom de
	 *  cap element de l'enumeració {@link DocIdentitat.TIPUS}
	 *  
	 */
	public void setTipusDI(String nomTipus) {
		this.tipusDI = nomTipus == null || nomTipus.isEmpty()?
				null : DocIdentitat.TIPUS.valueOf(nomTipus);
	}

	/**
	 * Accessor de lectura de l'atribut numDI
	 * @return el valor de l'atribut <code>numDI</code>
	 */
	public String getNumDI() {
		return this.numDI;
	}

	/**
	 * Accessor d'escriptura a l'atribut numDI
	 * @param numDI el nou valor de l'atribut numDI
	 */
	public void setNumDI(String numDI) {
		this.numDI = numDI;
	}

	/**
	 * Accessor de lectura de l'atribut nomUsuari
	 * @return el valor de l'atribut <code>nomUsuari</code>
	 */
	public String getNomUsuari() {
		return this.nomUsuari;
	}

	/**
	 * Accessor d'escriptura a l'atribut nomUsuari
	 * @param nomUsuari el nou valor de l'atribut nomUsuari
	 */
	public void setNomUsuari(String nomUsuari) {
		this.nomUsuari = nomUsuari;
	}

	/**
	 * Accessor de lectura de l'atribut patroNomComplet
	 * @return el valor de l'atribut <code>patroNomComplet</code>
	 */
	public String getPatroNomComplet() {
		return this.patroNomComplet;
	}

	/**
	 * Accessor d'escriptura a l'atribut patroNomComplet
	 * @param patroNomComplet el nou valor de l'atribut patroNomComplet
	 */
	public void setPatroNomComplet(String patroNomComplet) {
		this.patroNomComplet = patroNomComplet;
	}

	/* Dona la sentència JPQL que selecciona els clients que compleixen
	 * les condicions del filtre
	 * 
	 * @see edu.uoc.tcfj2ee_gatd.controlador.FiltreAbs#getJPQL()
	 */
	@Override
	public String getJPQL() {
		this.netejaJPQL();
		this.posaSelect("SELECT c FROM Client c");
		
		// No es mostren els clients de baixa
		this.posaAnd("c.dataBaixa IS NULL");
		
		this.posaAnd("c.di.tipus", this.tipusDI);
		this.posaAnd("c.di.numero", this.numDI);
		this.posaAnd("c.nomUsuari", this.nomUsuari);
		this.posaAndLike("c.contacte.nom", this.patroNomComplet);
		
		this.getLogger().finer("FiltreClients#getJPQL() ->" + Util.LS + 
				this.getSentencia());			
		
		return this.getSentencia();
	}


}
