/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.controlador.usuaris;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.faces.bean.ManagedBean;
import javax.faces.bean.SessionScoped;

/**
 * Historial de pàgines d'una sessió, per a assegurar el bon funcionament
 * del botó "tornar". Descarta les peticions HTTP successives a la mateixa pàgina.
 * <p>
 * Opera sota la premissa de que les pàgines o bé son llistats sens paràmetres
 * "query string", o son fitxes d'objecte individuals, diferenciades per la 
 * "query string".
 * </p>
 * @author Jordi Tejero
 */
@ManagedBean(name = "historial")
@SessionScoped
public class HistorialSessio implements Serializable {


	/**
	 * Versió de serialització
	 */
	private static final long serialVersionUID = 1L;

	/**
	 * La llista de les pàgines del servidor visitades per l'usuari de la
	 * sessió, sense comptar les repeticions
	 */
	private final List<Pagina> pagines;
	
	/**
	 * El logger
	 */
	private final Logger logger;
	
	
	/**
	 * Constructor
	 */
	public HistorialSessio() {
		this.logger = Logger.getLogger(this.getClass().getName());
		this.pagines = new ArrayList<Pagina>();
	}
	
	
	
	/**
	 * Accessor de lectura de l'atribut logger
	 * @return el valor de l'atribut logger
	 */
	private Logger getLogger() {
		return this.logger;
	}



	/**
	 * Afegir una pàgina a l'historial. Si coincideix amb la darrera
	 * introduïda, no fa res
	 * @param via el "path" de la nova pàgina
	 * @param qs la "query string" de la nova pàgina
	 */
	void avança(String via, String qs) {
		Pagina pag = new Pagina(via, qs);
		this.getLogger().finer(this.getClass().getSimpleName() +
			"#avança(" + pag.getURI() + ")");
		
		if(pag.repeteix(this.getPrevia())) {
			// detectat retrocés
			this.pagines.remove(this.pagines.size() - 1);			
			this.getLogger().finer("Retrocedit. Pàgs: " + this.pagines.size());
		}
		else if(!pag.repeteix(this.getDarrera())) {
			this.pagines.add(pag);
			this.getLogger().finer("Avançat. Pàgs: " + this.pagines.size());
		}
		else
			this.getLogger().finer("Repetició. Pàgs: " + this.pagines.size());
		
	}
	
	/**
	 * Dona la pàgina penúltima de l'historial
	 * @return la pàgina penúltima, o <code>null</code> si no existeix
	 */
	private Pagina getPrevia() {
		if(this.pagines.size() < 2)
			return null;
		return this.pagines.get(this.pagines.size() - 2);
	}
	
	/**
	 * Dona la URI de la pàgina prèviament visitada de l'historial
	 * @return la URI de la pàgina prèvia, o <code>null</code> si no existeix
	 */
	String getURIPrevia() {
		Pagina pag = this.getPrevia();
		return pag == null? null : pag.getURI();
	}



	/**
	 * Dona la darrera pàgina de l'historial
	 * @return la darrera pàgina, o <code>null</code> si l'historial és buit
	 */
	private Pagina getDarrera() {
		return this.pagines.isEmpty()? null 
			: this.pagines.get(this.pagines.size() - 1);
	}
	
	/**
	 * Dona la URI de la darrera pàgina de l'historial
	 * @return la URI de la darrera pàgina, o <code>null</code> si l'historial
	 *  es buit.
	 */
	String getURIDarrera() {
		Pagina pag = this.getDarrera();
		return pag == null? null : pag.getURI();
	}
	
	/**
	 * Representa una pàgina de l'historial, caracteritzada per un "path" i
	 * una "query string".
	 * @author Jordi Tejero
	 *
	 */
	private static class Pagina {
		/**
		 * "Path" de la pàgina
		 */
		private final String via;
		/**
		 * "Query string" de la pàgina
		 */
		private final String qs;
		
		/**
		 * Constructor
		 * @param via el "path"
		 * @param qs la "query string"
		 */
		Pagina(String via, String qs) {
			this.via = via;
			this.qs = qs;
		}
		
		
		/**
		 * Indica quan <code>this</code> es considera com a repetició d'una 
		 * pàgina precedent.
		 * <p>
		 * Reflecteix el fet de que si una nova pàgina té igual "path" que la
		 * actual però no té "query string", llavors és una recàrrega.
		 * </p>
		 * @param prec la pàgina precedent
		 * @return <code>true</code> sols si es dona un qualsevol d'aquest casos:
		 * <ol>
		 * 	<li><code>this.via == null</code></li>
		 *  <li><code>this.via == prec.via</code> i <code>this.qs == nova.qs</code></li>
		 *  <li><code>this.via == prec.via</code> i <code>this.via == null</code></li>
		 * </ol>
		 */
		boolean repeteix(Pagina prec) {
			if(prec == null)
				return false;
			
			return this.via == null 
			|| (this.via.equals(prec.via) && this.qs == null) 			
	        || (this.via.equals(prec.via) && this.qs.equals(prec.qs))
			;
		}
		
		/**
		 * URI de la pàgina representada
		 * @return [this.via]?[this.qs], o [this.via] si <code>this.qs == null</code>
		 */
		String getURI() {
			return this.qs == null? this.via : this.via + "?" + this.qs;
		}
	}
	
}
