package edu.uoc.tcfj2ee_gatd.model.atencio.entitats;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import javax.persistence.Entity;
import javax.persistence.ManyToOne;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.OneToMany;
import javax.persistence.PostLoad;
import javax.persistence.Transient;

import edu.uoc.tcfj2ee_gatd.model.comu.Util;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Empleat;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Peticio;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Seu;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Adreça;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Contacte;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.DocIdentitat;

/**
 * Un empleat amb perfil de tècnic
 *
 */
@Entity
@NamedQueries({
	@NamedQuery(name = "tecnicsActiusSeu", 
	query = "SELECT t FROM Tecnic t WHERE t.seu.codi = :codiSeu AND t.actiu = true")	
	})
public class Tecnic extends Empleat {

	/**
     * Indica si el tècnic és en situació d'activitat.<br/>
     * D'altra cas, no pot tenir peticions
     */
    private boolean actiu;

    /**
     * La seu a la que pertany el tècnic
     */
    @ManyToOne(optional = false)
    private Seu seu;
    
    /**
	 * Peticions de les que el tècnic és responsable.
	 * Estratègia de càrrega: <i>lazy</i>
	 */
    @OneToMany(mappedBy = "tecnic")    
	private Collection<Peticio> peticions = new LinkedList<Peticio>();
    
	
    /**
     * Les peticions en estat ASSIGNADA o ACCEPTADA del tècnic.
     * Si en te alguna d'acceptada, serà a l'index 0.
     */
    @Transient
    private List<Peticio> pendents;
    
	/**
	 * Constructor sense paràmetres, obligatori
	 */
	public Tecnic() {
		super();	
		this.seu = new Seu();
	}

	/**
	 * Constructor. 
	 * @param seu
	 * @param nomUsuari
	 * @param clauAccess
	 * @param di
	 * @param contacte
	 * @param adreça
	 * @param codiIdioma
	 * @param numSS
	 * @param actiu
	 */
	public Tecnic(
			Seu seu,
			String nomUsuari, String clauAccess, DocIdentitat di,
			Contacte contacte, Adreça adreça, String codiIdioma, String numSS,
			boolean actiu) {
		super(nomUsuari, clauAccess, di, contacte, adreça, codiIdioma, numSS);
		this.lliga(seu);
		this.actiu = actiu;
	}

	/**
	 * Funció de neteja, a executar cada cop que l'entitat es carrega
	 * des del proveïdor de persistència.
	 * <p>
	 * Esborra la llista de pendents. 
	 * </p>
	 */
	@PostLoad
	private void neteja() {
		//this.getLogger().finer(this.getClass().getSimpleName() + "#neteja()");
	    this.pendents = null;
	}
	
    /**
	 * Accessor de lectura a l'indicador d'activitat
	 * @return el valor de l'atribut <code>actiu</code>
	 */
	public boolean isActiu() {
		return this.actiu;
	}

	/**
	 * Accessor d'escriptura a l'indicador d'activitat
	 * @param actiu el nou valor de l'atribut <code>actiu</code>
	 */
	public void setActiu(boolean actiu) {
		this.actiu = actiu;
	}
	
	/**
	 * Afegeix una petició al tècnic, sense afectar al seu estat.
	 * Força l'anul·lació de la llista de pendents, que s'haurà de regenerar.
	 * 
	 * @param p la petició
	 */
	public void posat(Peticio p) {
		this.peticions.add(p);
		this.neteja();
	}
	
	/**
	 * Dona la seqüència de totes les peticions del tècnic.
	 * @return iterable que recorre totes les peticions
	 */
	public Iterable<Peticio> getPeticions() {
		return this.peticions;
	}

	/**
	 * Accessor de lectura la seu del tècnic
	 * @return el valor de l'atribut <code>seu</code>
	 */
	public Seu getSeu() {
		return this.seu;
	}

	/**
	 * Estableix la seu del tècnic. Bidireccional, actualitza també els enllaços
	 * dels objectes <code>Seu</code> actual i del nou.
	 * @param seu la nova seu del tècnic
	 */
	public void setSeu(Seu seu) {
		this.getLogger().finer(this.getClass().getSimpleName() 
			+ "#setSeu(" + seu.getBreu() + ")");
		this.lliga(seu);
	}
	
	/**
	 * Enllaça el tècnic amb la seva seu, bidireccionalment. El tècnic
	 * es baixa a la seu on era i alta a la nova
	 * @param s la nova seu del tècnic
	 */
	private void lliga(Seu s) {
		// Si el tècnic s'està construint, pot no tenir encara seu
		if(this.seu != null)
			this.seu.getTecnics().remove(this);
		this.seu = s;
		s.getTecnics().add(this);
	}
	
	/**
	 * Dona la petició acceptada del tècnic. Si cal, la cerca
	 * @return la petició acceptada, o <code>null</code> si no en té cap
	 */
	public Peticio getAcceptada() {		
	    // si té una petició acceptada, és a l'index 0 de la llista
		return 
			this.getLPendents().isEmpty() 
			|| !this.getLPendents().get(0).isAcceptada()? 
					null : this.getLPendents().get(0);
	}
	
	/**
	 * Estableix la petició acceptada pel tècnic, modificant-li l'estat.
	 * <p>
	 * <b>pre</b>
	 * </p>
	 * <p>
	 * <b>post</b> la petició és a la posició 0 de la llista de pendents 
	 * i és en estat ACCEPTADA.
	 * </p>
	 * 
	 * @param peti la nova petició acceptada
	 */
	public void accepta(Peticio peti) {
		assert peti.getTecnic() != null;
		
		this.getLPendents().remove(peti);
		
		peti.setEstat(Peticio.Estat.ACCEPTADA);
		this.getLPendents().add(0, peti);
	}

	/**
	 * Afegeix una petició a les del tècnic, modificant-li l'estat
	 * @param peti la nova petició acceptada
	 */
	public void assignat(Peticio peti) {
		peti.setEstat(Peticio.Estat.ASSIGNADA);
        this.peticions.add(peti);		
		this.getLPendents().add(peti);
	}	
	

	/**
	 * Retira totes les peticions pendents (estat ASSIGNADA o ACCEPTADA) de
	 * la col·lecció de peticions del tècnic. Anul·la la llista de peticions
	 * pendents.
	 * @return la llista prèvia de peticions pendents del tècnic
	 */
	public Collection<Peticio> retiratPendents() {
		Collection<Peticio> retorn = this.getPendents();
		this.peticions.removeAll(this.getPendents());
		this.neteja();
		
		this.getLogger().finer("Tecnic#retiraPendents(). " + 
				this.getBreu() +
				" Retirat " + 
			retorn.size() + " peticions pendents, en queden " +
			this.getPendents().size());
		
		return retorn;
		
	}
	
	/**
	 * Publica les peticions pendents, com una col·lecció. Si cal, la genera.
	 * <p>
	 * Si el tècnic té alguna petició acceptada, és a la posició 0 de la
	 * llista.
	 * </p>	  
	 * @return la col·lecció de les peticions pendents del tècnic
	 */
	public Collection<Peticio> getPendents() {
	   return this.getLPendents();
	}
	
	/**
	 * Dona el número de peticions assignades al tècnic
	 * @return el número de peticions pendents, tret si una és acceptada,
	 *  que torna el número de pendents - 1.
	 */
	public int getNumAssignades() {
		return this.getAcceptada() == null?
			this.getPendents().size() : this.getPendents().size() - 1
			;
	}

	/**
	 * Accés "mandrós" (<i>lazzy</i>) a la llista de les peticions pendents.
	 * Si no està creada, la crea.
	 * <p>
	 * Si el tècnic té alguna petició acceptada, és a la posició 0 de la
	 * llista.
	 * </p>
	 * @return la llista de les peticions pendents del tècnic
	 */
	private List<Peticio> getLPendents() {
	    if(this.pendents == null) {	        
	    	this.pendents = new LinkedList<Peticio>();
	    	for(Peticio p: this.peticions) {
		      if(p.isAcceptada())
		         this.pendents.add(0, p);
		      else if(p.isAssignada())
		          this.pendents.add(p);
	    	}	    	
	    }
		return this.pendents;
	}




	/*  Representació textual de l'objecte, per a depuració i proves.
	 * 
	 * @see edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Usuari#toString()
	 */
	@Override
	public String toString() {
		return super.toString() + Util.LS +
		"Seu: " + this.seu + ", actiu: " + Util.fmt(this.actiu) + ", " +
		"peticions: " + this.peticions.size()
		;
	}

}