/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.model.comu;

import java.util.Date;
import java.util.logging.Logger;

import javax.persistence.Column;
import javax.persistence.MappedSuperclass;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;
import javax.persistence.Transient;
import javax.persistence.Version;


/**
 * Forma general d'una classe d'entitat, que gestiona alguns
 * atributs comuns a totes elles:
 * <p>
 * Data d'alta, data de baixa i versió pel control conflictes
 * d'actualizació amb bloqueig optimista.
 * 
 * </p>
 * @author Jordi Tejero
 *
 */
@MappedSuperclass
public abstract class Entitat {

	
	/**
	 * Data i hora de creació de la entitat, establerta pel
	 * constructor
	 */
	@Temporal(TemporalType.TIMESTAMP)
	@Column(nullable = false)
	private Date dataAlta;
	
	/**
	 * Data i hora en que l'entatit va donar de baixa, o <code>null</code>
	 * si encara no ho ha estat.
	 */	
	@Temporal(TemporalType.TIMESTAMP)
	private Date dataBaixa;
	
	/**
	 * Número de versió gestionat autònomament pel proveïdor de
	 * persistència. Permet detectar actualitzacions conflictives.
	 */
	@Version
	private long versio;

	/**
	 * Logger 
	 */
	@Transient
	private final Logger logger;	
	

	/**
	 * Constructor. Estableix el logger
	 */
	public Entitat() {
		this.logger = Logger.getLogger(this.getClass().getName());
		/*
		this.logger.finest("Creat objecte entitat " +
			"[" + this.getClass().getSimpleName() + "]");*/

	}	
	

	/**
	 * Estableix l'observador de l'entitat
	 * @param observador l'observador. Pot ser <code>null</code>.
	 */
	/*
	protected void setObservador(ObservadorEntitat observador) {
		this.observador = observador;
		
	}
	*/

	/**
	 * Modifica la data d'alta.
	 * @param dataAlta el nou valor de l'atribut dataAlta
	 */
	public void setDataAlta(Date dataAlta) {
		this.dataAlta = dataAlta;
	}

	
	/**
	 * Modifica la data d'alta
	 * @return el valor de l'atribut <code>dataAlta</code>
	 */
	public Date getDataAlta() {
		return this.dataAlta;
	}

	/**
	 * Estableix la data de baixa de l'entitat
	 * @param dataBaixa el nou valor de l'atribut dataBaixa
	 */
	public void setDataBaixa(Date dataBaixa) {
		this.dataBaixa = dataBaixa;
	}

	/**
	 * Accessor de lectura a la data de baixa
	 * @return el valor de l'atribut <code>dataBaixa</code>
	 */
	public Date getDataBaixa() {
		return this.dataBaixa;
	}

	/**
	 * Indica si l'entitat ha estat donada de baixa
	 * @return <code>true</code> si l'entitat ha estat donada
	 *  de baixa.
	 */
	public boolean isBaixa() {
		return this.dataBaixa != null;
	}
	
	/**
	 * Marca l'entitat com a donada de baixa, establint al camp de data
	 * corresponent, la data actual.
	 */
	public void baixa() {
		this.dataBaixa = new Date();
	}
	
	/**
	 * Marca l'entitat com a estat d'alta, establint els camps de
	 * data corresponents
	 */
	public void alta() {
		this.dataAlta = new Date();
		this.dataBaixa = null;
	}

	/**
	 * Indica si l'entitat ha esta donada d'alta
	 * @return <code>true</code> si ho ha estat, encara que hagi estat donada 
	 *  també de baixa
	 */
	public boolean isAlta() {
		return this.dataAlta != null;
	}
	
	/**
     * Dona la versió de control de modifiació
     * @return el valor de l'atribut <code>versio</code>
     */
    public long getVersio() {
        return this.versio;
    }

    /**
     * Estableix la versió de control de modificació.
     * <p>
     * Necessari per a detectar quan una entitat ha estat modificada per
     * un altra usuari, mentre era editada (conflicte d'edicions).
     * </p>
     * @param versio el nou valor de l'atribut versio
     */
    public void setVersio(long versio) {
        this.versio = versio;

    }

    /**
	 * Publica el logger als descendents
	 * @return el valor de l'atribut <code>logger</code>
	 */
	protected Logger getLogger() {
		return this.logger;
	}
	
	
	/* Representació textual de la entitat, per a depuració i
	 * proves. No forma part de la lògica de negoci ni de
	 * la interfície.
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return 
		"[" + this.getClass().getSimpleName() + "]. " + 
		"Data alta: " + Util.fmt(this.dataAlta) + ". " +
		"Data baixa: " + Util.fmt(this.dataBaixa) + ". " +
		"Versio: " + this.versio
		;
	}
	
	
}
