/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.model.peticions;

import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;

import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.persistence.OptimisticLockException;
import javax.persistence.Query;

import edu.uoc.tcfj2ee_gatd.controlador.atencio.ConstructorEMail;
import edu.uoc.tcfj2ee_gatd.model.atencio.GATDExcepcioAtencio;
import edu.uoc.tcfj2ee_gatd.model.atencio.GestorEMail;
import edu.uoc.tcfj2ee_gatd.model.atencio.IConstructorEMail;
import edu.uoc.tcfj2ee_gatd.model.atencio.entitats.Tecnic;
import edu.uoc.tcfj2ee_gatd.model.comu.EJBGestor;
import edu.uoc.tcfj2ee_gatd.model.comu.Util;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Client;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Peticio;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Seu;
import edu.uoc.tcfj2ee_gatd.model.usuaris.GestorUsuaris;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.DistrictePostal;

/**
 * Session Bean que implementa la lògica de negoci del package
 * @author Jordi Tejero
 *
 */
@Stateless
public class GestorPeticions extends EJBGestor {

	/**
	 * Gestor en el que deleguem les operacions sobre els
	 * usuaris
	 */
	@EJB
	private GestorUsuaris gUsuaris;
	
	/**
	 * Gestor de correu electrònic que envia els missatges als tècnics
	 */
	@EJB
	private GestorEMail gEMail;

	/**
	 * El constructor dels correus electrònics que rebran els tècnics
	 */
	private final IConstructorEMail constructorEMail;
	
    /**
     * Constructor
     */
    public GestorPeticions() {
        super();
        
      /*assert this.getLogger() != null;
        this.getLogger().finer(this.getClass().getSimpleName() 
        		+ "#GestorPeticions()");*/
        
        this.constructorEMail = new ConstructorEMail();
    }

	/**
     * Dona la llista de totes les peticions
     * @return la llista de les peticions
     */
    @SuppressWarnings("unchecked")
	public List<Peticio> llistaPeticions() {
    	this.getLogger().finer("GATD. Llista de totes les peticions");
        Query query = this.getEM()
         	.createNamedQuery("totesPeticions");
         return query.getResultList();
    }
    
    /**
     * Dona la llista de totes les peticions d'un client
     * @param cl el client
     * @return la llista de les peticions
     * @throws GATDExcepcioPeticions si <code>cl</code> es <code>null</code>
     */
    @SuppressWarnings("unchecked")
    public List<Peticio> llistaPeticionsClient(Client cl) 
    throws GATDExcepcioPeticions {
    	if(cl == null)
    		throw new GATDExcepcioPeticions("peti_excp_NullClie");
    	this.getLogger().finer(
    		"GATD. Llista de les peticions del client " + cl.getNomUsuari());
         Query query = this.getEM().createNamedQuery("totesPeticionsClient");
         query.setParameter("nomUsu", cl.getNomUsuari());
         return query.getResultList();
    }

	/**
	 * Dona la llista amb les peticions d'un tècnic que estan pendents
	 * @param nomUsu el nom d'usuari del tècnic
	 * @return la llista de les peticions
	 */
	@SuppressWarnings("unchecked")
	public List<Peticio> llistaPetisPendentsTecnic(String nomUsu) {
    	assert nomUsu != null;
    	this.getLogger().finer(
    		"GATD. Llista de les peticions del tècnic " + nomUsu);
         Query query = this.getEM().createNamedQuery("peticionsPendentsTecnic");
         query.setParameter("nomUsu", nomUsu);
         return query.getResultList();
	}


	/**
	 * Alta d'una nova petició.
	 * <p>
	 * Valida del districte postal i li assigna un tècnic
	 * </p>
	 * 
	 * @param peti la petició
	 *  
	 * @throws GATDExcepcioAtencio si l'adreça no es correcta o el 
	 *  codi postal no és a cap seu
	 */
	public void novaPeticio(Peticio peti)
	throws GATDExcepcioPeticions {
		String cp = peti.getAdreça().getDp().getCodi();
		DistrictePostal dp = this.getEM().find(DistrictePostal.class, cp);
		if(dp == null || dp.getSeu() == null)
			throw new GATDExcepcioPeticions("peti_excp_CPForani", cp);
		peti.getAdreça().setDp(dp);
		
		this.assignaTecnicPeticio(peti);
		
		
	}		
	
	/**
	 * Assigna un tècnic a una petició, la desa al gestor de persistència
	 * i envia un e-mail al tènic.
	 * <p>
	 * <b>pre</b> el districte postal de la petició correspon a una de les
	 *  seus de l'empressa i dita seu té tècnics actius
	 * </p>
	 * <b>post</b> la petició ha estat assignada a un dels tècnics de la seu
	 * corresponent al districte postal i és en estat ASSIGNADA
	 * @param peti la petició
	 * @throws GATDExcepcioPeticions cas d'error al enviar el correu electrònic
	 *  al tècnic
	 */
	@SuppressWarnings("unchecked")
	public void assignaTecnicPeticio(Peticio peti) 
	throws GATDExcepcioPeticions {
		assert peti != null;
		
		Seu seu = peti.getAdreça().getDp().getSeu();
		assert seu != null;
		
		// Cerca dels tècnics actius a la seu
		List<Tecnic> lTCandidats;
		Query query = this.getEM()
			.createNamedQuery("tecnicsActiusSeu");
		query.setParameter("codiSeu", seu.getCodi());
		lTCandidats = query.getResultList();
		
		assert lTCandidats.size() > 0;
		
		/* Ens quedem sols amb els tècnics amb el nombre mínim de peticions
		 * pendents
		 */
		int min = lTCandidats.get(0).getPendents().size();
		List<Tecnic> lTDispos = new LinkedList<Tecnic>();
		for(Tecnic t : lTCandidats) {
			assert t.isActiu();
			int m = t.getPendents().size();
			if(m < min) {
				// Nou mínim
				min = m;
				lTDispos.clear();
			}
			if(m == min)
				lTDispos.add(t);				
		}
		
		lTCandidats = null;
		
		this.getLogger().finer(
			"GestorPeticions#assignaTecnicPeticio" +
			"(" + peti.getReferencia() + "). " + Util.LS +	
			"Trobat " + lTDispos.size() + " tècnics " +
			"disponibles a la seu " + seu.getNom() + " amb " + min +
			" peticions");
		
		assert lTDispos.size() > 0;
		
		/* Assignem la petició al primer tècnic lliure
		 */
		
		Tecnic responsable = lTDispos.get(0);
		
		this.assigna(responsable, peti);
		
		assert this.gEMail != null;
		
		try {			
			/* Si la petició és nova, no tindrà número de referència fins
			 * que no la guardem
			 */
			this.guarda(peti);
			this.gEMail.envia(this.constructorEMail.getMsgAssignada(
				responsable, peti));
		} catch (GATDExcepcioAtencio e) {
			throw new GATDExcepcioPeticions(e);
		}
		
		this.getLogger().finer(
				"GestorPeticions#assignaTecnicPeticio" +
				"(" + peti.getReferencia() + "). " + Util.LS +	
				"Assignat el tècnic: " + Util.LS +
				peti.getTecnic().getBreu());
			
	}
	
	/**
	 * Assigna un tècnic a una petició.
	 * Estableix l'associació bidireccional i deixa la petició en estat
	 * assignada.
	 *  
	 * @param t el tècnic que serà responsable de la petició
	 * @param p la petició
	 */
	private void assigna(Tecnic t, Peticio p) {
		assert t.getSeu().equals(p.getAdreça().getDp().getSeu());
		p.setTecnic(t);
		t.assignat(p);
	}
	
	/**
	 * Acceptació d'una petició per part d'un tècnic
	 * @param el nom d'usuari del tècnic
	 * @param p la petició
	 * @throws GATDExcepcioPeticions cas de conflicte d'edicions
	 */
	public void accepta(String nomTecnic, Peticio p) throws GATDExcepcioPeticions {
		assert nomTecnic != null;
		Tecnic t = this.gUsuaris.cerca(Tecnic.class, nomTecnic);
		assert t.isActiu();
		assert p != null;
		
		this.getLogger().finer(this.getClass().getSimpleName() + 
			"#accepta(" + t.getBreu() + ", " + p.getBreu() + ")");
		
		t.accepta(p);
		p.setTecnic(t);
		this.guarda(p);
		this.getEM().merge(t);
		
	}
	
	/**
	 * Dona una petició a partir del seu número de referència
	 * @param referencia el número de referència
	 * @return la <code>Peticio</code> que té tal número, o <code>null</code> 
	 * si cap no el té.
	 */
	public Peticio cerca(Long referencia) {
		return this.getEM().find(Peticio.class, referencia);
	}
	
	/**
	 * Cerca una seu pel seu codi
	 * @param codiSeu el codi de la seu que es cerca
	 * @return la <code>Seu</code> que te tal codi, o <code>null</code>
	 *  si cap no el té
	 */
	public Seu cercaSeu(String codiSeu) {
		return this.getEM().find(Seu.class, codiSeu);
	}

    /**
     * Guarda una petició al sistema. Si es nova, l'afegeix. Si no, la
     * modifica.
     * @param peti la petició.
     * @return la pròpia petició <code>peti</code> si ha estat donada d'alta,
     *  o la seva forma administrada si ha estat modificada
     * @throws GATDExcepcioAtencio cas de conflicte d'edicions o d'un error
     *  inesperat
     */
    public Peticio guarda(Peticio peti) throws GATDExcepcioPeticions {
    	Peticio retorn;
    	
        this.getLogger().finer(
             "GestorPeticions#guarda(" + peti.getReferencia() + ")");
        //this.getLogger().finer(peti.toString());
        
        /* 
         * No guardem una queixa que no ha arribat ha estar d'alta
         */
        if(peti.getQueixa() != null && !peti.getQueixa().isAlta())
        	peti.setQueixa(null);
        
        try {
	        if(!peti.isAlta()) {
	            retorn = this.alta(peti);
	            this.getLogger().finer(
	                 "GestorPeticions#guarda(" + retorn.getReferencia() + "). " +
	                 "Dada d'alta");            
	        }
	        else { 
	            retorn = this.getEM().merge(peti);
	            this.getLogger().finer(
	                "GestorPeticions#guarda(" + peti.getReferencia() + "). " +
	                "Merged");            
	        }
	       
	        this.getEM().flush();
        }
        catch(OptimisticLockException e) {
        	this.getLogger().log(Level.WARNING, 
    				"GestorPeticions#guarda(). Colisió d'edicions: " + Util.LS + 
    				peti,  e);        	
        	throw new GATDExcepcioPeticions("peti_excp_ConflicteEdicio", 
        			peti.getReferencia());
        }
        catch(Throwable e) {
        	this.getLogger().log(Level.WARNING, 
    				"GestorPeticions#guarda(). Error inesperat: " + Util.LS + 
    				peti,  e);        	
        	throw new GATDExcepcioPeticions("err_excp_Inesperat", 
        			e.toString());
        }
        
        return retorn;
    }

	/**
	 * Dona una petició com a finalitzada per part d'un tècnic.
	 * <p> 
	 * Verifica que el tècnic sigui el responsable de la petició i que
	 * el temps invertit que s'hi ha posat sigui més gran que zero.
	 * </p>
	 * @param nomTecnic el nom d'usuari del tècnic que finalitza la
	 *  petició.
	 * @param p la petició a finalitzar
	 * 
	 * @throws GATDExcepcioPeticions si falla alguna de les verificacions
	 *   anteriors o si no es pot guardar la petició finalitzada al gestor de 
	 *   persistència. 
	 *   Per exemple, perquè hi ha hagut un conflicte d'edicions,
	 */
	public void finalitza(String nomTecnic, Peticio p) 
	throws GATDExcepcioPeticions {
		assert p != null;		
		Tecnic t = this.gUsuaris.cerca(Tecnic.class, nomTecnic);		
		assert t != null;
		
		this.getLogger().finer(this.getClass().getSimpleName() 
			+ "#finalitza(" + p.getBreu() + " en " + p.getTemps() + ")");
		
		if(p.getTemps() <= 0.0)			
			throw new GATDExcepcioPeticions("peti_excp_EntraTemps");
				
		p.finalitzat();
		this.guarda(p);

		/* Que el tècnic responsable de la petició no sigui el que vol 
		 * finalitzar-la, pot ser degut a un conflicte d'edicions.
		 * Per això deixem la verificació pel final. 
		 */
		if(!t.equals(p.getTecnic()))
			throw new GATDExcepcioPeticions("peti_excp_DiscrepaTecnic", 
					p.getReferencia(), t.getNomUsuari());
		
	}

	/**
	 * Anul·la una petició, verificant abans que estigui pendent
	 * @param p la petició a anul·lar
	 * @throws GATDExcepcioPeticions si la petició no està pendent o
	 *  si es produeix un conflicte d'edicions al guardar-la.
	 */
	public void anulla(Peticio p) throws GATDExcepcioPeticions {
		assert p != null;
		
		if(!p.isPendent())
			throw new GATDExcepcioPeticions("peti_excp_NoAnulable",
				p.getReferencia(), p.getEstat());
		p.anullat();
		this.guarda(p);
		
		/*this.getLogger().finer(this.getClass().getSimpleName() + 
			"#anulla(). Petició " + p.getBreu() + " estat " + p.getEstat() + 
			", data baixa " + p.getDataBaixa());*/
	}
	

    
}
