/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.model.peticions.entitats;

import java.util.Date;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.OneToOne;

import edu.uoc.tcfj2ee_gatd.configuracio.GestorConfiguracio;
import edu.uoc.tcfj2ee_gatd.model.atencio.entitats.Tecnic;
import edu.uoc.tcfj2ee_gatd.model.comu.Entitat;
import edu.uoc.tcfj2ee_gatd.model.comu.Util;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Adreça;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Contacte;


/**
 * Una petició d'assitencia feta per un client
 * @author Jordi Tejero
 *
 */
@Entity
@NamedQueries({
	@NamedQuery(name = "totesPeticions", query = "SELECT p FROM Peticio p"),
	@NamedQuery(name = "totesPeticionsClient", 
		query = "SELECT p FROM Peticio p WHERE p.client.nomUsuari = :nomUsu"),
	@NamedQuery(name = "peticionsPendentsTecnic", 
		query = 
		"SELECT p FROM Peticio p WHERE p.tecnic.nomUsuari = :nomUsu " +
		"AND p.estat <> edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Peticio.Estat.FINALITZADA " +
		"AND p.estat <> edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Peticio.Estat.ANULLADA " +
		"ORDER BY p.estat ASC, p.referencia ASC"
		
		)
	})
public class Peticio extends Entitat {

	
	/**
	 * Els estats en que pot ser una petició
	 *
	 */
	public enum Estat {
		REBUDA, ASSIGNADA, ACCEPTADA, FINALITZADA, ANULLADA
	}
	
	/**
	 * Número únic de referència de la petició, donat pel sistema.
	 */
	@Id @GeneratedValue
	private Long referencia;
	
	/**
	 * Client que ha fet la petició
	 */
	@ManyToOne(optional = false)
	private Client client;

	/**
	 * Estat de la petició. Inicialment, <code>REBUDA</code>.
	 * A la taula de persistència, serà una cadena.
	 */
	@Enumerated(EnumType.STRING)
	@Column(nullable = false, length = 20)
	private Estat estat;
	
	/**
	 * Adreça del domicili de la petició
	 */
	@OneToOne(cascade = CascadeType.ALL)
	@JoinColumn(nullable = false)
	private Adreça adreça;
	
	/**
	 * Descripció de l'actuació demanada
	 */
	@Column(length = 500)
	private String descripcio;

	/**
	 * Dades de la persona de contacte
	 */
	@Embedded	
	private Contacte contacte;

	/**
	 * Tècnic responsable de la petició, o <code>null</code> si
	 * encara és en estat <code>REBUDA</code> i no ha estat assignada
	 * a cap.
	 */
	@ManyToOne
	private Tecnic tecnic; 
	
	/**
	 * Temps invertit en finalitzar la petició, o 0 si no és en
	 * estat FINALITZADA.
	 */
	private float temps;

	/**
	 * Possible queixa sobre la petició, feta pel peticionari
	 * o per un controlador. 
	 * <p>Si no se n'ha fet cap, serà <code>null</code>, però un
	 * cop feta, ja no ho serà, encara que s'anul·li.
	 */
	@OneToOne(cascade = CascadeType.ALL)
	private Queixa queixa;

	/**
	 * Constructor sense paràmetres, obligatori. Estableix l'estat inicial 
	 * <code>REBUDA</code>
	 */	
	public Peticio() {
		
		this.estat = Estat.REBUDA;
		this.adreça = new Adreça();
		this.contacte = new Contacte();
		
		this.temps = 0;
		
	}
	
	
	/**
	 * Constructor amb tots els atributs
	 * 
	 * @param client el client que fa la petició
	 * @param adreça l'adreça del domicili on cal anar
	 * @param contacte dades de la persona de contacte
	 * @param descripcio descripció de l'actuació demanada.
	 * @param tecnic tècnic responsable. Pot ser <code>null</code> 
	 * @param estat estat de la petició
	 * @param temps temps invertit en finalitzar-la. 
	 *  0 si no està finalitzada.
	 */
	public Peticio(
			Date dataAlta, Date dataBaixa,
			Client client, Adreça adreça, Contacte contacte,
			String descripcio,
			Tecnic tecnic, Estat estat, float temps) {
		this();
		
		assert tecnic == null || tecnic.getSeu().equals(adreça.getDp().getSeu());
		
		this.setDataAlta(dataAlta);
		this.setDataBaixa(dataBaixa);
		this.lliga(client);
		this.lliga(tecnic);
		this.estat = estat;
		this.adreça = adreça;
		this.descripcio = descripcio;
		this.contacte = contacte;
		this.temps = temps;
	}
	
	/**
	 * Crea una petició per a un client, amb el domicili d'actuació i
	 * dades de contacte, copiats dels del client.
	 * 
	 * @param client el client pel que es crea l'actuació.
	 */
	public Peticio(Client client) {
		this();
		this.lliga(client);
		this.adreça = new Adreça(client.getAdreça());
		this.contacte = new Contacte(client.getContacte());		
	}

	/**
	 * Lliga la petició amb el client que la fa, bidireccionalment
	 * @param cl el client que fa la petició
	 */
	private void lliga(Client cl) {
		this.client = cl;
		cl.getPeticions().add(this);
	}
	
	/**
	 * Lliga la petició amb el tècnic responsable, bidireccionalment
	 * @param t el tècnic responsable
	 */
	private void lliga(Tecnic t) {
		this.tecnic = t;
		t.posat(this);
	}
	
	/**
	 * Accessor de lectura de l'atribut estat
	 * @return el valor de l'atribut <code>estat</code>
	 */
	public Estat getEstat() {
		return this.estat;
	}

	/**
	 * Accessor d'escriptura a l'atribut estat
	 * @param estat el nou valor de l'atribut estat
	 */
	public void setEstat(Estat estat) {
		this.estat = estat;
	}

	/**
	 * Publica el número de referència de la petició
	 * @return el valor de l'atribut <code>referencia</code>
	 */
	public Long getReferencia() {
		return this.referencia;
	}

	/**
	 * Estableix el número de referència.
	 * @param referencia el número de referència
	 */
	public void setReferencia(Long referencia) {
	    this.referencia = referencia;
	}
	
	/**
	 * Publica el client que ha fet la petició
	 * @return el valor de l'atribut <code>client</code>
	 */
	public Client getClient() {
		return this.client;
	}

	/**
	 * Estableix el client que fa la petició, lligant-lo bidireccionalemnt
	 * @param cl el client peticionari
	 */
	public void setClient(Client cl) {
		this.lliga(cl);
		
	}

	/**
	 * Publica l'adreça del domicili on cal actuar
	 * @return el valor de l'atribut <code>adreça</code>
	 */
	public Adreça getAdreça() {
		return this.adreça;
	}

	/**
	 * Accessor d'escriptura a l'atribut adreça
	 * @param adreça el nou valor de l'atribut adreça
	 */
	public void setAdreça(Adreça adreça) {
		this.adreça = adreça;
	}



	/**
	 * Accessor de lectura de l'atribut contacte
	 * @return el valor de l'atribut <code>contacte</code>
	 */
	public Contacte getContacte() {
		return this.contacte;
	}
	
	/**
	 * Accessor d'escriptura a l'atribut contacte
	 * @param contacte el nou valor de l'atribut contacte
	 */
	public void setContacte(Contacte contacte) {
		this.contacte = contacte;
	}



	/**
	 * Publica el tècnic responsable de la petició. <code>null</code>
	 * si no ha estat encara assignada
	 * @return el valor de l'atribut <code>tecnic</code>
	 */
	public Tecnic getTecnic() {
		return this.tecnic;
	}
	
	/**
	 * Estableix el tècnic responsable de la petició
	 * @param tecnic el nou valor de l'atribut tecnic
	 */
	public void setTecnic(Tecnic tecnic) {
		this.tecnic = tecnic;
	}


	/**
	 * Accés a la possible queixa
	 * @return el valor de l'atribut <code>queixa.</code> <code>null</code> si no
	 * se n'ha fet cap
	 */
	public Queixa getQueixa() {
		return this.queixa;
	}



	/**
	 * Estableix una queixa per a la petició, lligant-la bidireccionalmente
	 * @param queixa la queixa, o <code>null</code> per a deslligar l'actual
	 */
	public void setQueixa(Queixa queixa) {
		if(this.queixa != null)
			this.queixa.setPeticio(null);
		this.queixa = queixa;
		if(this.queixa != null)
			this.queixa.setPeticio(this);
	}



	/**
	 * Accessor de lectura de l'atribut descripcio
	 * @return el valor de l'atribut <code>descripcio</code>
	 */
	public String getDescripcio() {
		return this.descripcio;
	}



	/**
	 * Accessor d'escriptura a l'atribut descripcio
	 * @param descripcio el nou valor de l'atribut descripcio
	 */
	public void setDescripcio(String descripcio) {
		this.descripcio = descripcio;
	}



	/**
	 * Publica el temps invertit en la finalització de la petició.
	 * @return el valor de l'atribut <code>temps</code>
	 */
	public float getTemps() {
		/*this.getLogger().finer(this.getClass().getSimpleName() 
			+ "#getTemps() -> " + this.temps);*/
		return this.temps;
	}


	/**
	 * Estableix el temps invertit en la finalització de la petició.
	 * @param temps el nou valor de l'atribut temps
	 */
	public void setTemps(float temps) {
	/*	this.getLogger().finer(this.getClass().getSimpleName() 
			+ "#setTemps(" + temps + ")");*/
		this.temps = temps;
	}


	/**
	 * Atribut derivat que indica si la petició està pendent o ja ha estat
	 * tancada
	 * @return <code>false</code> si la petició és en estat 
	 *  <code>ANULLADA</code> o FINALITZADA, <code>true</code> si es en un altre 
	 *  estat.
	 */
	public boolean isPendent() {
		return this.estat != Estat.ANULLADA && this.estat != Estat.FINALITZADA;
	}
	
	/**
	 * Indica si la petició es en estat ASSIGNADA
     * @return true sols si ho està
     */
    public boolean isAssignada() {
        return this.estat.equals(Peticio.Estat.ASSIGNADA);
    }

    /**
     * Indica si la petició es en estat ACCEPTADA
     * @return true sols si ho està
     */
    public boolean isAcceptada() {
        return this.estat.equals(Peticio.Estat.ACCEPTADA);
    }   	

	/**
	 * Anul·la la petició, marcant-la com a baixa
	 */
	public void anullat() {
		this.estat = Estat.ANULLADA;
		this.baixa();
	}
	
	/**
	 * Finalitza la petició, modificant el seu estat i marcant-la  
	 * com a baixa.
	 * 
	 */
	public void finalitzat() {
		
		this.estat = Estat.FINALITZADA;
		this.baixa();
	}
	
	/**
	 * Indica si la petició ha excedit el temps configurat sense ser
	 * acceptada
	 * @return <code>true</code> si i sols si la petició és en estat
	 *  ASSIGNADA i han passat més hores que el llindar d'alerta configurat 
	 */
	public boolean isAlertaAtencio() {
		return this.isAssignada() && 
		this.passa(GestorConfiguracio.getConfig().getLlindarAlertaAtencio());
	}
	
	/**
	 * Indica si la petició ha excedit el temps configurat sense ser
	 * tancada
	 * @return <code>true</code> si i sols si la petició està pendent
	 *  i han passat més hores que el llindar d'alerta configurat 
	 */
	public boolean isAlertaFi() {
		assert GestorConfiguracio.getConfig() != null;
		return this.isPendent() && 
		this.passa(GestorConfiguracio.getConfig().getLlindarAlertaFi());
	}	
	
	/**
	 * Indica si la petició té activa alguna alerta
	 * @return <code>true</code> si la petició té una queixa, si ha excedit el 
	 *  llindar de temps sense ser atesa o ha excedit el llindar de temps sense 
	 *  ser finalitzada.
	 */
	public boolean isAlerta() {
		return this.queixa != null || this.isAlertaAtencio() || this.isAlertaFi();
	}
	
	/**
	 * Indica si la petició ha superat el llindar de temps per a ser atesa
	 * o el de ser resolta
	 * @return <code>true</code> si ha superat qualsevol dels dos llindars
	 *  de temps
	 */
	public boolean isDemorada() {
		return this.isAlertaAtencio() || this.isAlertaFi();
	}
	
	/**
	 * Indica si la petició té una queixa que no ha estat anul·lada
	 * @return true si i sols sí <code>this.queixa != null</code> i 
	 *  <code>!this.queixa.isBaixa()</code>. 
	 */
	public boolean isQueixada() {
		return this.queixa != null && !this.queixa.isBaixa();
	}
	
	/**
	 * Indica si han passat més d'un cert nombre d'hores des de la data
	 * d'alta de la petició
	 * @param hores el nombre d'hores
	 * @return <code>true</code> si i sols si la data actual es posterior
	 *  a la data d'alta més el nombre d'hores indicat. En particular,
	 *  si la petició no està d'alta, retorna <code>false</code>.
	 */
	private boolean passa(float hores) {
		if(this.getDataAlta() == null)
			return false;
		
		Date llindar = new Date(this.getDataAlta().getTime() + 
				(long)(hores * 1000 * 60 * 60));
		return new Date().compareTo(llindar) > 0;
	}
	
	/**
	 * Auxiliar per al "logging". Dona una representació abreujada de la
	 * petició.
	 * @return cadena amb la classe i el número de referència
	 */
	public String getBreu() {
		return "[" + this.getClass().getSimpleName() + "] " + this.referencia; 
	}


	/*
	 * Dues peticions son iguales si les dues son no <code>null</code>
	 * i tenen el mateix número de referència
	 * @param p objecte amb el que comparem
	 * @return <code>true</code> si i sols si <code>p</code> és de classe 
	 * <code>Peticio</code> i té la mateixa referència que <code>this</code>.
	 */
	@Override
	public boolean equals(Object p) {
		if(p == null)
			return false;
		if(!(p instanceof Peticio))
			return false;
		Peticio pt = (Peticio)p;
		return this.referencia != null 
		&& this.referencia.equals(pt.getReferencia());
	}
	
	/* Codi de dispersió, compatible amb equals()
	 * @return el número de referència
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.referencia == null? 0 : this.referencia.intValue();
	}
	
	/*  Representació textual de l'objecte, per a depuració i proves.
	 * @see edu.uoc.tcfj2ee_gatd.model.comu.Entitat#toString()
	 */
	@Override
	public String toString() {
		return 
		"-------- ini petició " + Util.LS +
		super.toString() + Util.LS +
		"Referència: " + this.referencia + ", " +
		"estat: " + this.estat + ", temps: " + this.temps + Util.LS +
		"Client: " + this.client + Util.LS +
		"Anar a: " + this.adreça + Util.LS +
		"Contactar amb: " + this.contacte + Util.LS +
		"Responsable: " + this.tecnic + Util.LS +
		"Queixa: " + this.queixa + Util.LS +
		"-------- fi petició"
		;
	}
	
}
