/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.model.usuaris;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.persistence.OptimisticLockException;
import javax.persistence.Query;

import edu.uoc.tcfj2ee_gatd.model.atencio.entitats.Tecnic;
import edu.uoc.tcfj2ee_gatd.model.comu.EJBGestor;
import edu.uoc.tcfj2ee_gatd.model.comu.GATDExcepcioExecucio;
import edu.uoc.tcfj2ee_gatd.model.comu.Util;
import edu.uoc.tcfj2ee_gatd.model.peticions.GATDExcepcioPeticions;
import edu.uoc.tcfj2ee_gatd.model.peticions.GestorPeticions;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Peticio;
import edu.uoc.tcfj2ee_gatd.model.peticions.entitats.Seu;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.DistrictePostal;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.DocIdentitat;
import edu.uoc.tcfj2ee_gatd.model.usuaris.entitats.Usuari;





/**
 * Session Bean que implementa la lògica de negoci del package
 * @author Jordi Tejero
 *
 */
@Stateless
public class GestorUsuaris extends EJBGestor {

	/**
	 * Alta d'un districte postal donat pel seu codi.
	 * No l'assigna a cap seu
	 * @param codi el codi postal
	 * @return el nou districte postal
	 */
	public DistrictePostal altaDP(String codi) {
		return this.alta(new DistrictePostal(codi));
	}
	
	/**
	 * Gestor de peticions al que es demanen certes operacions 
	 */
	@EJB
	private GestorPeticions gPeticions;

	/**
	 * Troba un districte postal pel seu codi. Si no existeix, el
	 * dona d'alta sense assignar-li cap seu
	 * @param codi el codi postal
	 * @return el districte postal
	 * @throws GATDExcepcioExecucio si <code>codi</code> es <code>null</code>
	 */
	public DistrictePostal donamDP(String codi) throws GATDExcepcioExecucio {
		if(codi == null)
			throw new GATDExcepcioExecucio("usua_excp_BuscaCPNull");
		DistrictePostal result = this.getEM().find(DistrictePostal.class, codi);
		if(result == null)
			result = this.altaDP(codi);
		return result;
	}
	
	/**
	 * Cerca un usuari pel seu nom d'usuari, amb exigència d'existència
	 * @param nomUsuari el nom d'usuari de l'usuari
	 * @return l'usuari amb el nom cercat
	 * @throws GATDExcepcioUsuaris si no hi ha cap usuari amb el nom cercat 
	 */
	public <U extends Usuari> U troba(Class<U> k, String nomUsuari) 
	throws GATDExcepcioUsuaris {
	    
	    U u = this.getEM().find(k, nomUsuari);
	    
	    if(u == null)
	        throw new GATDExcepcioUsuaris("usua_excp_NomNoExisteix", nomUsuari);
	    
	    return u;
	}

	/**
	 * Dona la llista de tots els districtes postals
	 * @return la llista del districtes
	 */
	@SuppressWarnings("unchecked")
	public List<DistrictePostal> llistaDP() {
		 Query query = this.getEM().createNamedQuery("totsDP");
	     return query.getResultList();
	}

	/**
	 * Persistència d'un nou usuari. Si ja existia un amb el mateix nom,
	 * excepció.
	 * @param nou usuari a persistir
	 * @return el mateix usuari <code>nou</code>
	 * @throws GATDExcepcioUsuaris si ja existeix un usuari amb el
	 *  mateix nom que <code>nou</code>, o amb el mateix document
	 *  d'identitat
	 */
	public <U extends Usuari> U alta(U nou) throws GATDExcepcioUsuaris {
		assert nou != null;
		
		this.getLogger().finer(this.getClass().getSimpleName() + 
			"#alta(" + nou.getBreu() + ")");
		
		if(this.getEM().find(Usuari.class, nou.getNomUsuari()) != null)
			throw new GATDExcepcioUsuaris("usua_excp_NomExisteix", 
				nou.getNomUsuari());
		
		
		// lliguem adreça amb codi postal
		nou.getAdreça().setDp(this.donamDP(nou.getAdreça().getDp().getCodi()));
		// persistim
		super.alta(nou);
		
		/*
		 * Sembla que aquesta comprovació s'ha de fer després de persistir
		 * l'usuari. Si no, es produeix una excepció "During synchronization a 
		 * new object was found..." 
		 */
		List<Usuari> lUDI = this.llistaPerDI(nou.getDi());
		this.getLogger().finer(this.getClass().getSimpleName() + 
			"#alta(). Usuaris amb igual DI: " + lUDI.size());
		
		if(lUDI.size() > 1)
			throw new GATDExcepcioUsuaris("usua_excp_DIExisteix");
		

		
		return nou;
	}

	/**
	 * Cerca els usuaris que tenen un document d'identitat donat.
	 * <p>
	 * No hi hauria d'haver més d'un, però el model de dades no ho
	 * garanteix.
	 * </p>
	 * @param di el document d'identitat
	 * @return la llista d'usuaris amb el document d'identitat indicat, pot
	 *  ser que buida.
	 */
	@SuppressWarnings("unchecked")
	private List<Usuari> llistaPerDI(DocIdentitat di) {
		
		/*this.getLogger().finer(this.getClass().getSimpleName() + 
			"#llistaPerDI(" + di.toString() + ")");*/
		
		Query query = this.getEM().createNamedQuery("usuarisDocIdentitat");
		query.setParameter("di", di);
		
		List llista = query.getResultList();
		
		/* this.getLogger().finer(this.getClass().getSimpleName() + 
			"#llistaPerDI(). Trobat " + llista.size());*/
		
		return llista;
		
	}
	
	/**
	 * Dona de baixa un usuari. Estableix la seva data de baixa i el guarda
	 * al gestor de persistència
	 * @param u l'usuari a donar de baixa
	 * @throws GATDExcepcioUsuaris cas d'error
	 */
	public void baixa(Usuari u) throws GATDExcepcioUsuaris {
		u.baixa();
		this.actualitza(u);
	}

	
	/**
	 * Cerca un usuari d'una classe dada, pel seu nom
	 * @param <U> tipus de la classe de l'usuari
	 * @param cla la classe de l'usuari que es cerca
	 * @param nomUsuari el nom d'usuari, N. Pot ser <code>null</code>
	 * @return l'usuari U de tipus E, amb U.nomUsuari == N, o 
	 * <code>null</code> si <code>nomUsuari == null</code> o si cap usuari de 
	 *  classe E té nom N
	 */
	public <U extends Usuari> U cerca(Class<U> cla, String nomUsuari) {
		U u = nomUsuari == null?
			null : this.getEM().find(cla, nomUsuari);
		
		return u;
	}
	


	/**
	 * Cerca una seu pel seu codi
	 * @param codiSeu el codi de la seu a cercar
	 * @return la seu amb el codi cercat, o <code>null</code> si cap seu no
	 *  el té.
	 */
	public Seu cercaSeu(String codiSeu) {
		return this.getEM().find(Seu.class, codiSeu);
	}

	/**
	 * Actualitza un usuari al gestor de persistència.
	 * <p>
	 * Si l'usuari ha estat marcat com a baixa, li esborrem la clau
	 * d'accés per a que no pugui entrar.
	 * </p> 
	 * <p>Si l'usuari ha estat
	 * actualitzar al gestor de persistència després d'haver estat recuperat,
	 * es produirà una excepció.
	 * </p>
	 * @param usuari l'usuari a actualitzar
	 * @return L'usuari de tipus E, administrat
	 * @throws GATDExcepcioUsuaris si s'ha produït un conflicte d'edicions o
	 *  s'ha intentat donar a l'usuari un document d'identitat que es d'un
	 *  altre
	 */
	public <E extends Usuari> E actualitza(E usuari) throws GATDExcepcioUsuaris {
		E retorn;
	

		// Verificar que no hi ha cap altre usuari amb el mateix doc. d'identitat
		List<Usuari> lUDI = this.llistaPerDI(usuari.getDi());
		this.getLogger().finer(this.getClass().getSimpleName() + 
			"#actualitza(). Usuaris amb igual DI: " + lUDI.size());
		
		for(Usuari u : lUDI)
			if(!u.equals(usuari))
				throw new GATDExcepcioUsuaris("usua_excp_DIExisteix");
		
		if(usuari.isBaixa())
			usuari.setClauAcces(null);

		try {
			retorn = this.getEM().merge(usuari);			
		} 
		catch(OptimisticLockException e) {
			throw new GATDExcepcioUsuaris("usua_excp_ConflicteEdicio", 
				usuari.getNomUsuari());
		}
				
		return retorn;
		
	}

	
	
	/**
	 * Implementació del cas d'ús "Modifica Situació Tècnics".
	 * <p>
	 * Commuta la situació d'activitat d'una col·leció de tècnics.
	 * <p>
	 * Verifica que tots son de la mateixa seu i que un cop fet el canvi,
	 * la seu tindrà al menys un tècnic actiu.
	 * </p>
	 * <p>
	 * Les peticions pendents dels tècnics de la col·lecció que passin a situació
	 * d'inactivitat, seran reassignades.
	 * </p>
	 * @param marcats col·lecció dels tècnics marcats per a 
	 *  commutar la  situació.
	 *   
	 * @throws GATDExcepcioUsuaris en qualsevol d'aquest casos:
	 * <ol> 
	 * <li>La col·lecció de tècnics es buida</li>
	 * <li>Hi ha tècnics de més d'una seu</li> 
	 * <li>Algun tècnic no està d'alta</li>
	 * <li>Si la seu comuna es quedaria sense tècnics actius</li>
	 * </ol> 
	 * La operació queda marcada per a fer "rollback".
	 */
	private void modificaSituacioTecnics(Collection<Tecnic> marcats) 
	throws GATDExcepcioUsuaris {

		this.getLogger().finer("GestorUsuaris#modificaSituacioTecnics()");

		// peticions a reassignar
		Collection<Peticio> petis = new LinkedList<Peticio>();
		
		if(marcats.isEmpty())
			throw new GATDExcepcioUsuaris("usua_excp_CapTecnic");		
		
		Iterator<Tecnic> itt = marcats.iterator();
		assert itt.hasNext();
		
		Tecnic t = itt.next();
		final Seu seu = t.getSeu();
		
		boolean fi = false;
		while(!fi) {
			this.getLogger().finer("GestorUsuaris#modificaSituacioTecnics(). " +
				t.getBreu() + " seu " + t.getSeu().getNom());
			if(!seu.equals(t.getSeu()))
				throw new GATDExcepcioUsuaris("usua_excp_SeusDiferents", 
					t.getNomUsuari(), t.getSeu().getNom(), seu.getNom());
			if(!t.isAlta() || t.isBaixa())
				throw new GATDExcepcioUsuaris("usua_excp_TecnicNoAlta", 
						t.getNomUsuari());
			
			t.setActiu(!t.isActiu());
			// retirem les peticions als tècnics que queden inactius
			if(!t.isActiu())
				petis.addAll(t.retiratPendents());
			this.getLogger().finer("GestorUsuaris#modificaSituacioTecnics(). " +
					t.getBreu() + " queda actiu: " + t.isActiu());
						
			this.getEM().merge(t);
			
			fi = !itt.hasNext();
			if(!fi)
				t = itt.next();
		}
		
		if(this.comptaTecnicsActius(seu) == 0)
			throw new GATDExcepcioUsuaris("usua_excp_SeuSenseTActius", 
					seu.getNom());
		
		// Tornar a assignar un tècnic a les peticions
		for(Peticio p : petis) {
			assert p.isPendent();
			try {
				this.gPeticions.assignaTecnicPeticio(p);
			} catch (GATDExcepcioPeticions e) {
				throw new GATDExcepcioUsuaris(e);
			}
		}
	}

	/**
	 * Implementació del cas d'ús "Modifica Situació Tècnics".
	 * <p>
	 * Commuta la situació d'activitat d'una col·leció de tècnics, donada
	 * pels seus noms d'usuari.
	 * <p>
	 * Verifica que tots son de la mateixa seu i que un cop fet el canvi,
	 * la seu tindrà al menys un tècnic actiu.
	 * </p>
	 * <p>
	 * Les peticions pendents dels tècnics de la col·lecció que passin a situació
	 * d'inactivitat, seran reassignades.
	 * </p>
	 * <p>
	 * Per a evitar actuar sobre tècnics que puguin haver estat modificats
	 * (per exemple, donats de baixa o activats/desactivats) després d'haver-se
	 * carregat el llistat, es comprova el número de versió d'entitat de
	 * cada un.
	 * </p>
	 * 	 
	 * @param marcats array amb els noms dels tècnics marcats per a 
	 *  commutar la  situació.
	 * @param versions array amb els números de versió d'entitat que esperem
	 *  tinguin els tècnics 
	 *  
	 * @throws GATDExcepcioUsuaris en qualsevol d'aquest casos:
	 * 
	 * <ol> 
	 * <li>L'<i>array</i> de noms de tècnics es buit</li>
	 * <li>Algun dels noms no es correspon amb cap tècnic</li>
	 * <li>Algun dels tècnics no té el número de versió d'entitat esperat 
	 *  (conflicte d'edicions)</li>
	 * <li>Hi ha tècnics de més d'una seu</li> 
	 * <li>Algun tècnic no està d'alta</li>
	 * <li>Si la seu comuna es quedaria sense tècnics actius</li>
	 * </ol> 
	 * La operació queda marcada per a fer "rollback".
	 */
	public void modificaSituacioTecnics(
			String[] marcats, long[] versions) 
	throws GATDExcepcioUsuaris {

		this.getLogger().finer("GestorUsuaris#modificaSituacioTecnics(marcats[], versions[])");
				
		Collection<Tecnic> tMarcats = null;
		if(marcats == null)
			tMarcats = Collections.emptyList();
		else
			tMarcats = this.trobaUsuaris(Arrays.asList(marcats));
		
		// Cerquem un possible conflicte d'edicions abans de res
		int i = 0;
		for(Tecnic t : tMarcats)
			if(t.getVersio() != versions[i++])
				throw new GATDExcepcioUsuaris("usua_excp_ConflicteEdicio", 
					t.getNomUsuari());
		
		this.modificaSituacioTecnics(tMarcats);
	}
	
	
	/**
	 * Cerca una col·leció d'usuaris, donats pel seu nom
	 * @param <U> el tipus dels usuaris que cerca
	 * @param noms col·lecció no buida amb els noms d'usuari
	 * @return una llista d'usuaris de tipus U
	 *  
	 * @throws GATDExcepcioUsuaris si es troba un nombre diferents d'usuaris
	 *  del que es demana
	 */
	@SuppressWarnings({ "unchecked" })
	private <U extends Usuari> List<U> trobaUsuaris(Collection<String> noms) 
	throws GATDExcepcioUsuaris {
		assert !noms.isEmpty();
		
		StringBuilder jpql = new StringBuilder();
		for(String nom : noms) {
			if(jpql.length() == 0) 
				jpql.append("SELECT u FROM Usuari u").append(Util.LS)
				.append("WHERE ");
			else
				jpql.append(Util.LS).append("OR ");
			jpql.append("u.nomUsuari = '").append(nom).append("'");				
		}
		
		this.getLogger().finer("GestorUsuaris#trobaUsuaris. jpql = " + Util.LS +
			jpql.toString());
		
		List<U> retorn = this.getEM().createQuery(jpql.toString()).getResultList();
		
		if(retorn.size() != noms.size())
			throw new GATDExcepcioUsuaris("usua_excp_TrobaDiscrepancia", 
					retorn.size(), noms.size());
		this.getLogger().finer("GestorUsuaris#trobaUsuaris. Trobat " +
			retorn.size() + " usuaris");
		
		return retorn;
	}

	/**
	 * Compta el nombre de tècnics actius que hi ha a una seu
	 * @param seu la seu
	 * @return el nombre de tècnics d'alta en situació d'activitat
	 */
	public int comptaTecnicsActius(Seu seu) {
		int retorn = 0;
		for(Tecnic t : seu.getTecnics())
			if(t.isAlta() && t.isActiu())
				++retorn;
		
		return retorn;
	}

	/**
	 * Dona la seqüència de totes les seus d'alta
	 * @return Iterable que recorre totes les seus que no son de baixa
	 */
	public Iterable<Seu> getSeus() {
		return this.llistaNominal("seusAlta");
	}
	

}
