/**
 * 
 */
package edu.uoc.tcfj2ee_gatd.model.usuaris.entitats;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.DiscriminatorColumn;
import javax.persistence.DiscriminatorType;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.Inheritance;
import javax.persistence.InheritanceType;
import javax.persistence.JoinColumn;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.OneToOne;

import edu.uoc.tcfj2ee_gatd.model.comu.Entitat;
import edu.uoc.tcfj2ee_gatd.model.comu.Util;


/**
 * Usuari amb les seves dades d'accés.
 * <p>
 * La taula de persistència és també la que conté les dades
 * per a autoritzar l'accés. La columna discriminatoria donarà el
 * role de l'usuari. 
 * </p>
 * <p>
 * Com que el nom de la taula i els de les columnes del nom d'usuari, clau
 * i role han de ser conegudes pel gestor del regne de seguretat (<i>realm</i>),
 * els donem explícitament
 * </p>
 * @author Jordi Tejero
 *
 */
@Entity
@Inheritance(strategy = InheritanceType.SINGLE_TABLE)
// Role = nom de la subclasse
@DiscriminatorColumn(name = "CLASSE", discriminatorType = DiscriminatorType.CHAR)
// Clau alternativa. NO FUNCIONA
// @Table(uniqueConstraints = {@UniqueConstraint(columnNames = {"di"})})
@NamedQueries({
	@NamedQuery(name = "usuarisDocIdentitat", 
		query = "SELECT u FROM Usuari u WHERE u.di = :di")		
	})
public class Usuari extends Entitat {
	
	/**
	 * Nom d'usuari al sistema
	 */
	@Id
	@Column(length = 20, name = "NOMUSUARI")
	private String nomUsuari;	
	/**
	 * Clau d'accés. Pot ser <code>null</code>, cas en que l'usuari no pot entrar
	 */
	@Column(nullable = true, length = 20, name = "CLAU")
	private String clauAcces;
	
	/**
	 * Document d'identitat de l'usuari
	 */
	@Embedded
	@Column(nullable = false)
	private DocIdentitat di;
	/**
	 * Dades per a contactar-hi. Inclouen el nom complet.
	 */
	@Embedded
	@Column(nullable = false)
	private Contacte contacte; 
		
	/**
	 * Adreça del usuari 
	 */
	@JoinColumn(nullable = false)
	@OneToOne(cascade = CascadeType.ALL)
	private Adreça adreça;

	/**
	 * Codi ISO del idioma que fa servir a la interfície
	 */
	@Column(nullable = false, length = 6)
	private String codiIdioma;
	
	/**
	 * Constructor sense paràmetres, obligatori.
	 * <p>Crea els objectes atribut, buits.
	 * </p>
	 */
	public Usuari() {
		super();
		this.di = new DocIdentitat();
		this.contacte = new Contacte();
		this.adreça = new Adreça();
	}
	
	/**
	 * Constructor
	 * @param nomUsuari
	 * @param clauAcces
	 * @param di
	 * @param contacte
	 * @param adreça
	 * @param codiIdioma El codi d'idioma per a la interfície. 
	 */
	protected Usuari(String nomUsuari, String clauAccess, 
			DocIdentitat di, Contacte contacte,
			Adreça adreça, String codiIdioma) {
		this();
		this.di = di;
		this.nomUsuari = nomUsuari;
		this.clauAcces = clauAccess;
		this.contacte = contacte;
		this.adreça = adreça;
		this.codiIdioma = codiIdioma;
	}


	/**
	 * Estableix el nom d'usuari. 
	 * @param nomUsuari el nou valor de l'atribut nomUsuari
	 */
	public void setNomUsuari(String nomUsuari) {
		this.nomUsuari = nomUsuari;
		
	}

	/**
	 * Accessor de lectura del nom d'usuari
	 * @return el valor de nomUsuari
	 */
	public String getNomUsuari() {
		return this.nomUsuari;
	}

	/**
	 * Accessor de lectura de l'atribut adreça
	 * @return el valor de l'atribut <code>adreça</code>
	 */
	public Adreça getAdreça() {
		return this.adreça;
	}


	/**
	 * Publica el codi ISO del idioma de la interfície amb l'usuari
	 * @return el valor de l'atribut <code>codiIdioma</code>
	 */
	public String getCodiIdioma() {
		return this.codiIdioma;
	}


	/**
	 * Modifica el codi ISO del idioma de la interfície amb l'usuari
	 * @param codiIdioma el nou valor de l'atribut [codiIdioma]
	 */
	public void setCodiIdioma(String codiIdioma) {
		this.codiIdioma = codiIdioma;
		// this.getLogger().finest("Usuari#setCodiIdioma(" + codiIdioma + ")"); 
	}

	/**
	 * Accessor de lectura de l'atribut contacte
	 * @return el valor de l'atribut <code>contacte</code>
	 */
	public Contacte getContacte() {
		return this.contacte;
	}


	/**
	 * Accessor d'escriptura a l'atribut clauAcces
	 * @param clauAcces el nou valor de l'atribut clauAcces. 
	 *  Pot ser <code>null</code>, per a evitar que l'usuari
	 *  pugui entrar.
	 */
	public void setClauAcces(String clauAccess) {
		this.clauAcces = clauAccess;
	}


	/**
	 * Accessor de lectura de clauAcces
	 * @return el valor de clauAcces
	 */
	public String getClauAcces() {
		return this.clauAcces;
	}


	/**
	 * Accessor de lectura del document d'identitat
	 * @return el document d'identitat
	 */
	public DocIdentitat getDi() {
		return this.di;
	}
	
	/* Dos usuaris son el mateix si tenen igual nom d'usuari, encara que
	 * siguin de subclasses diferents.
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object altre) {
		
		if(altre == null)
			return false;
		if(!(altre instanceof Usuari))
			return false;
		
		Usuari u = (Usuari) altre;		
		return this.nomUsuari != null && this.nomUsuari.equals(u.nomUsuari);
		
	}
	
	/* Codi de dispersió compatible amb "equals(...)".
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.nomUsuari == null? 0 : this.nomUsuari.hashCode();
	}
	
	/* Representació textual de l'objecte, per a depuració i proves
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return super.toString() + Util.LS +
		"Nom d'usuari: " + this.nomUsuari + ", " +
		"clau d'accés: " + this.clauAcces + Util.LS +
		this.di + Util.LS +
		this.contacte + Util.LS +
		this.adreça + Util.LS +
		"Idioma: " + this.codiIdioma
		;
				
	}
	
	/**
	 * Auxiliar per al "logging". Dona una representació abreujada de l'usuari
	 * @return cadena amb la classe i el nom de l'usuari
	 */
	public String getBreu() {
		return "[" + this.getClass().getSimpleName() + "] " + this.nomUsuari; 
	}
}
